import numpy as np

from napari._vispy.overlays.labels_polygon import VispyLabelsPolygonOverlay
from napari.components.overlays import LabelsPolygonOverlay
from napari.layers.labels._labels_key_bindings import complete_polygon
from napari.utils.interactions import (
    mouse_move_callbacks,
    mouse_press_callbacks,
)


def test_vispy_labels_polygon_overlay(make_napari_viewer):
    viewer = make_napari_viewer()

    labels_polygon = LabelsPolygonOverlay()

    data = np.zeros((50, 50), dtype=int)
    layer = viewer.add_labels(data, opacity=0.5)

    vispy_labels_polygon = VispyLabelsPolygonOverlay(
        layer=layer, overlay=labels_polygon
    )

    assert vispy_labels_polygon._polygon.color.alpha == 0.5

    labels_polygon.points = []
    assert not vispy_labels_polygon._line.visible
    assert not vispy_labels_polygon._polygon.visible

    labels_polygon.points = [(0, 0), (1, 1)]
    assert vispy_labels_polygon._line.visible
    assert not vispy_labels_polygon._polygon.visible
    assert np.allclose(
        vispy_labels_polygon._line.color[:3], layer._selected_color[:3]
    )

    labels_polygon.points = [(0, 0), (1, 1), (0, 3)]
    assert not vispy_labels_polygon._line.visible
    assert vispy_labels_polygon._polygon.visible

    layer.selected_label = layer.colormap.background_value
    assert vispy_labels_polygon._polygon.color.is_blank


def test_labels_drawing_with_polygons(MouseEvent, make_napari_viewer):
    """Test polygon painting."""
    np.random.seed(0)

    data = np.zeros((3, 15, 15), dtype=np.int32)
    viewer = make_napari_viewer()
    layer = viewer.add_labels(data)

    layer.mode = 'polygon'
    layer.selected_label = 1

    # Place some random points and then cancel them all
    for _ in range(5):
        position = (0,) + tuple(np.random.randint(20, size=2))
        event = MouseEvent(
            type='mouse_press',
            button=1,
            position=position,
            dims_displayed=[1, 2],
        )
        mouse_press_callbacks(layer, event)

    # Cancel all the points
    for _ in range(5):
        event = MouseEvent(
            type='mouse_press',
            button=2,
            position=(0, 0, 0),
            dims_displayed=(1, 2),
        )
        mouse_press_callbacks(layer, event)

    assert np.array_equiv(data[0, :], 0)

    # Draw a rectangle (the latest two points will be cancelled)
    points = [
        (1, 1, 1),
        (1, 1, 10),
        (1, 10, 10),
        (1, 10, 1),
        (1, 12, 0),
        (1, 0, 0),
    ]
    for position in points:
        event = MouseEvent(
            type='mouse_move',
            button=None,
            position=(1,) + tuple(np.random.randint(20, size=2)),
            dims_displayed=(1, 2),
        )
        mouse_move_callbacks(layer, event)

        event = MouseEvent(
            type='mouse_press',
            button=1,
            position=position,
            dims_displayed=[1, 2],
        )
        mouse_press_callbacks(layer, event)

    # Cancel the latest two points
    for _ in range(2):
        event = MouseEvent(
            type='mouse_press',
            button=2,
            position=(1, 5, 1),
            dims_displayed=(1, 2),
        )
        mouse_press_callbacks(layer, event)

    # Finish drawing
    complete_polygon(layer)

    assert np.array_equiv(data[[0, 2], :], 0)
    assert np.array_equiv(data[1, 1:11, 1:11], 1)
    assert np.array_equiv(data[1, 0, :], 0)
    assert np.array_equiv(data[1, :, 0], 0)
    assert np.array_equiv(data[1, 11:, :], 0)
    assert np.array_equiv(data[1, :, 11:], 0)

    # Try to finish with an incomplete polygon
    for position in [(0, 1, 1)]:
        event = MouseEvent(
            type='mouse_press',
            button=1,
            position=position,
            dims_displayed=(1, 2),
        )
        mouse_press_callbacks(layer, event)

    # Finish drawing
    complete_polygon(layer)
    assert np.array_equiv(data[0, :], 0)



import numpy as np

from napari._tests.utils import skip_local_popups, skip_on_win_ci


def test_multiscale(make_napari_viewer):
    """Test rendering of multiscale data."""
    viewer = make_napari_viewer()

    shapes = [(4000, 3000), (2000, 1500), (1000, 750), (500, 375)]
    np.random.seed(0)
    data = [np.random.random(s) for s in shapes]
    _ = viewer.add_image(data, multiscale=True, contrast_limits=[0, 1])
    layer = viewer.layers[0]

    # Set canvas size to target amount
    viewer.window._qt_viewer.canvas.size = (800, 600)
    viewer.window._qt_viewer.canvas.on_draw(None)

    # Check that current level is first large enough to fill the canvas with
    # a greater than one pixel depth
    assert layer.data_level == 2

    # Check that full field of view is currently requested
    assert np.all(layer.corner_pixels[0] <= [0, 0])
    assert np.all(layer.corner_pixels[1] >= np.subtract(shapes[2], 1))

    # Test value at top left corner of image
    viewer.cursor.position = (0, 0)
    value = layer.get_value(viewer.cursor.position, world=True)
    np.testing.assert_allclose(value, (2, data[2][(0, 0)]))

    # Test value at bottom right corner of image
    viewer.cursor.position = (3995, 2995)
    value = layer.get_value(viewer.cursor.position, world=True)
    np.testing.assert_allclose(value, (2, data[2][(999, 749)]))

    # Test value outside image
    viewer.cursor.position = (4000, 3000)
    value = layer.get_value(viewer.cursor.position, world=True)
    assert value[1] is None


def test_3D_multiscale_image(make_napari_viewer):
    """Test rendering of 3D multiscale image uses lowest resolution."""
    viewer = make_napari_viewer()

    data = [np.random.random((128,) * 3), np.random.random((64,) * 3)]
    viewer.add_image(data)

    # Check that this doesn't crash.
    viewer.dims.ndisplay = 3

    # Set canvas size to target amount
    viewer.window._qt_viewer.canvas.size = (800, 600)
    viewer.window._qt_viewer.canvas.on_draw(None)

    # Check lowest resolution is used
    assert viewer.layers[0].data_level == 1


@skip_on_win_ci
@skip_local_popups
def test_multiscale_screenshot(make_napari_viewer):
    """Test rendering of multiscale data with screenshot."""
    viewer = make_napari_viewer(show=True)

    shapes = [(4000, 3000), (2000, 1500), (1000, 750), (500, 375)]
    data = [np.ones(s) for s in shapes]
    _ = viewer.add_image(data, multiscale=True, contrast_limits=[0, 1])

    # Set canvas size to target amount
    viewer.window._qt_viewer.canvas.view.canvas.size = (800, 600)

    screenshot = viewer.screenshot(canvas_only=True, flash=False)
    center_coord = np.round(np.array(screenshot.shape[:2]) / 2).astype(int)
    target_center = np.array([255, 255, 255, 255], dtype='uint8')
    target_edge = np.array([0, 0, 0, 255], dtype='uint8')
    screen_offset = 3  # Offset is needed as our screenshots have black borders

    np.testing.assert_allclose(screenshot[tuple(center_coord)], target_center)
    np.testing.assert_allclose(
        screenshot[screen_offset, screen_offset], target_edge
    )
    np.testing.assert_allclose(
        screenshot[-screen_offset, -screen_offset], target_edge
    )


@skip_on_win_ci
@skip_local_popups
def test_multiscale_screenshot_zoomed(make_napari_viewer):
    """Test rendering of multiscale data with screenshot after zoom."""
    viewer = make_napari_viewer(show=True)
    view = viewer.window._qt_viewer

    shapes = [(4000, 3000), (2000, 1500), (1000, 750), (500, 375)]
    data = [np.ones(s) for s in shapes]
    _ = viewer.add_image(data, multiscale=True, contrast_limits=[0, 1])

    # Set canvas size to target amount
    view.canvas.view.canvas.size = (800, 600)

    # Set zoom of camera to show highest resolution tile
    view.canvas.view.camera.rect = [1000, 1000, 200, 150]
    viewer.window._qt_viewer.canvas.on_draw(None)

    # Check that current level is bottom level of multiscale
    assert viewer.layers[0].data_level == 0

    screenshot = viewer.screenshot(canvas_only=True, flash=False)
    center_coord = np.round(np.array(screenshot.shape[:2]) / 2).astype(int)
    target_center = np.array([255, 255, 255, 255], dtype='uint8')
    screen_offset = 3  # Offset is needed as our screenshots have black borders

    # for whatever reason this is the only test where the border is 6px on hi DPI.
    # if the 6 by 6 corner is all black assume we have a 6px border.
    if not np.allclose(screenshot[:6, :6], np.array([0, 0, 0, 255])):
        screen_offset = 6  # Hi DPI
    np.testing.assert_allclose(screenshot[tuple(center_coord)], target_center)
    np.testing.assert_allclose(
        screenshot[screen_offset, screen_offset], target_center
    )
    np.testing.assert_allclose(
        screenshot[-screen_offset, -screen_offset], target_center
    )


@skip_on_win_ci
@skip_local_popups
def test_image_screenshot_zoomed(make_napari_viewer):
    """Test rendering of image data with screenshot after zoom."""
    viewer = make_napari_viewer(show=True)
    view = viewer.window._qt_viewer

    data = np.ones((4000, 3000))
    _ = viewer.add_image(data, multiscale=False, contrast_limits=[0, 1])

    # Set canvas size to target amount
    view.canvas.view.canvas.size = (800, 600)

    # Set zoom of camera to show highest resolution tile
    view.canvas.view.camera.rect = [1000, 1000, 200, 150]
    viewer.window._qt_viewer.canvas.on_draw(None)

    screenshot = viewer.screenshot(canvas_only=True, flash=False)
    center_coord = np.round(np.array(screenshot.shape[:2]) / 2).astype(int)
    target_center = np.array([255, 255, 255, 255], dtype='uint8')
    screen_offset = 3  # Offset is needed as our screenshots have black borders

    np.testing.assert_allclose(screenshot[tuple(center_coord)], target_center)
    np.testing.assert_allclose(
        screenshot[screen_offset, screen_offset], target_center
    )
    np.testing.assert_allclose(
        screenshot[-screen_offset, -screen_offset], target_center
    )


@skip_on_win_ci
@skip_local_popups
def test_multiscale_zoomed_out(make_napari_viewer):
    """See https://github.com/napari/napari/issues/4781"""
    # Need to show viewer to ensure that pixel_scale and physical_size
    # get set appropriately.
    viewer = make_napari_viewer(show=True)
    shapes = [(3200, 3200), (1600, 1600), (800, 800)]
    data = [np.zeros(s, dtype=np.uint8) for s in shapes]
    layer = viewer.add_image(data, multiscale=True)
    qt_viewer = viewer.window._qt_viewer
    # Canvas size is in screen pixels.
    qt_viewer.canvas.size = (1600, 1600)
    # The camera rect is (left, top, width, height) in scene coordinates.
    # In this case scene coordinates are the same as data/world coordinates
    # the layer is 2D and data-to-world is identity.
    # We pick a camera rect size that is much bigger than the data extent
    # to simulate being zoomed out in the viewer.
    camera_rect_size = 34000
    camera_rect_center = 1599.5
    camera_rect_start = camera_rect_center - (camera_rect_size / 2)
    qt_viewer.canvas.view.camera.rect = (
        camera_rect_start,
        camera_rect_start,
        camera_rect_size,
        camera_rect_size,
    )

    qt_viewer.canvas.on_draw(None)

    assert layer.data_level == 2


@skip_on_win_ci
@skip_local_popups
def test_5D_multiscale(make_napari_viewer):
    """Test 5D multiscale data."""
    # Show must be true to trigger multiscale draw and corner estimation
    viewer = make_napari_viewer(show=True)
    view = viewer.window._qt_viewer
    view.set_welcome_visible(False)
    shapes = [(1, 2, 5, 20, 20), (1, 2, 5, 10, 10), (1, 2, 5, 5, 5)]
    np.random.seed(0)
    data = [np.random.random(s) for s in shapes]
    layer = viewer.add_image(data, multiscale=True)
    assert layer.data == data
    assert layer.multiscale is True
    assert layer.ndim == len(shapes[0])


@skip_on_win_ci
@skip_local_popups
def test_multiscale_flipped_axes(make_napari_viewer):
    """Check rendering of multiscale images with negative scale values.

    See https://github.com/napari/napari/issues/3057
    """
    viewer = make_napari_viewer(show=True)

    shapes = [(4000, 3000), (2000, 1500), (1000, 750), (500, 375)]
    data = [np.ones(s) for s in shapes]
    # this used to crash, see issue #3057
    _ = viewer.add_image(
        data, multiscale=True, contrast_limits=[0, 1], scale=(-1, 1)
    )


@skip_on_win_ci
@skip_local_popups
def test_multiscale_rotated_image(make_napari_viewer):
    viewer = make_napari_viewer(show=True)
    sizes = [4000 // i for i in range(1, 5)]
    arrays = [np.zeros((size, size), dtype=np.uint8) for size in sizes]
    for arr in arrays:
        arr[:10, :10] = 255
        arr[-10:, -10:] = 255

    viewer.add_image(arrays, multiscale=True, rotate=44)
    screenshot_rgba = viewer.screenshot(canvas_only=True, flash=False)
    screenshot_rgb = screenshot_rgba[..., :3]
    assert np.any(
        screenshot_rgb
    )  # make sure there is at least one white pixel



import csv
import os
from pathlib import Path
from typing import NamedTuple
from uuid import uuid4

import dask.array as da
import imageio
import npe2
import numpy as np
import pytest
import tifffile
import zarr

from napari_builtins.io._read import (
    _guess_layer_type_from_column_names,
    _guess_zarr_path,
    csv_to_layer_data,
    magic_imread,
    read_csv,
)
from napari_builtins.io._write import write_csv


class ImageSpec(NamedTuple):
    shape: tuple[int, ...]
    dtype: str
    ext: str
    levels: int = 1


PNG = ImageSpec((10, 10), 'uint8', '.png')
PNG_RGB = ImageSpec((10, 10, 3), 'uint8', '.png')
PNG_RECT = ImageSpec((10, 15), 'uint8', '.png')
TIFF_2D = ImageSpec((15, 10), 'uint8', '.tif')
TIFF_3D = ImageSpec((2, 15, 10), 'uint8', '.tif')
ZARR1 = ImageSpec((10, 20, 20), 'uint8', '.zarr')


@pytest.fixture
def write_spec(tmp_path: Path):
    def writer(spec: ImageSpec):
        image = np.random.random(spec.shape).astype(spec.dtype)
        fname = tmp_path / f'{uuid4()}{spec.ext}'
        if spec.ext == '.tif':
            tifffile.imwrite(str(fname), image)
        elif spec.ext == '.zarr':
            fname.mkdir()
            z = zarr.open(store=str(fname), mode='a', shape=image.shape)
            z[:] = image
        else:
            imageio.imwrite(str(fname), image)
        return fname

    return writer


def test_no_files_raises(tmp_path):
    with pytest.raises(ValueError, match='No files found in'):
        magic_imread(tmp_path)


def test_guess_zarr_path():
    assert _guess_zarr_path('dataset.zarr')
    assert _guess_zarr_path('dataset.zarr/some/long/path')
    assert not _guess_zarr_path('data.tif')
    assert not _guess_zarr_path('no_zarr_suffix/data.png')


def test_zarr(tmp_path):
    image = np.random.random((10, 20, 20))
    data_path = str(tmp_path / 'data.zarr')
    z = zarr.open(store=data_path, mode='a', shape=image.shape)
    z[:] = image
    image_in = magic_imread([data_path])
    np.testing.assert_array_equal(image, image_in)


def test_zarr_nested(tmp_path):
    image = np.random.random((10, 20, 20))
    image_name = 'my_image'
    root_path = tmp_path / 'dataset.zarr'
    grp = zarr.open(store=str(root_path), mode='a')
    grp.create_dataset(image_name, data=image, shape=image.shape)

    image_in = magic_imread([str(root_path / image_name)])
    np.testing.assert_array_equal(image, image_in)


def test_zarr_with_unrelated_file(tmp_path):
    image = np.random.random((10, 20, 20))
    image_name = 'my_image'
    root_path = tmp_path / 'dataset.zarr'
    grp = zarr.open(store=str(root_path), mode='a')
    grp.create_dataset(image_name, data=image, shape=image.shape)

    txt_file_path = root_path / 'unrelated.txt'
    txt_file_path.touch()

    image_in = magic_imread([str(root_path)])
    np.testing.assert_array_equal(image, image_in[0])


def test_zarr_multiscale(tmp_path):
    multiscale = [
        np.random.random((20, 20)),
        np.random.random((10, 10)),
        np.random.random((5, 5)),
    ]
    fout = str(tmp_path / 'multiscale.zarr')

    root = zarr.open_group(fout, mode='a')
    for i in range(len(multiscale)):
        shape = 20 // 2**i
        z = root.create_dataset(str(i), shape=(shape,) * 2, dtype=np.float64)
        z[:] = multiscale[i]
    multiscale_in = magic_imread([fout])
    assert len(multiscale) == len(multiscale_in)
    for images, images_in in zip(multiscale, multiscale_in, strict=False):
        np.testing.assert_array_equal(images, images_in)


def test_write_csv(tmpdir):
    expected_filename = os.path.join(tmpdir, 'test.csv')
    column_names = ['column_1', 'column_2', 'column_3']
    expected_data = np.random.random((5, len(column_names)))

    # Write csv file
    write_csv(expected_filename, expected_data, column_names=column_names)
    assert os.path.exists(expected_filename)

    # Check csv file is as expected
    with open(expected_filename) as output_csv:
        csv.reader(output_csv, delimiter=',')
        for row_index, row in enumerate(output_csv):
            if row_index == 0:
                assert row == 'column_1,column_2,column_3\n'
            else:
                output_row_data = [float(i) for i in row.split(',')]
                np.testing.assert_allclose(
                    np.array(output_row_data), expected_data[row_index - 1]
                )


def test_read_csv(tmpdir):
    expected_filename = os.path.join(tmpdir, 'test.csv')
    column_names = ['column_1', 'column_2', 'column_3']
    expected_data = np.random.random((5, len(column_names)))

    # Write csv file
    write_csv(expected_filename, expected_data, column_names=column_names)
    assert os.path.exists(expected_filename)

    # Read csv file
    read_data, read_column_names, _ = read_csv(expected_filename)
    read_data = np.array(read_data).astype('float')
    np.testing.assert_allclose(expected_data, read_data)

    assert column_names == read_column_names


def test_guess_layer_type_from_column_names():
    points_names = ['index', 'axis-0', 'axis-1']
    assert _guess_layer_type_from_column_names(points_names) == 'points'

    shapes_names = ['index', 'shape-type', 'vertex-index', 'axis-0', 'axis-1']
    assert _guess_layer_type_from_column_names(shapes_names) == 'shapes'

    also_points_names = ['no-index', 'axis-0', 'axis-1']
    assert _guess_layer_type_from_column_names(also_points_names) == 'points'

    bad_names = ['no-index', 'no-axis-0', 'axis-1']
    assert _guess_layer_type_from_column_names(bad_names) is None


def test_read_csv_raises(tmp_path):
    """Test various exception raising circumstances with read_csv."""
    temp = tmp_path / 'points.csv'

    # test that points data is detected with require_type = None, any, points
    # but raises for other shape types.
    data = [['index', 'axis-0', 'axis-1']]
    data.extend(np.random.random((3, 3)).tolist())
    with open(temp, mode='w', newline='') as csvfile:
        csv.writer(csvfile).writerows(data)
    assert read_csv(temp, require_type=None)[2] == 'points'
    assert read_csv(temp, require_type='any')[2] == 'points'
    assert read_csv(temp, require_type='points')[2] == 'points'
    with pytest.raises(ValueError, match='not recognized as'):
        read_csv(temp, require_type='shapes')

    # test that unrecognized data is detected with require_type = None
    # but raises for specific shape types or "any"
    data = [['some', 'random', 'header']]
    data.extend(np.random.random((3, 3)).tolist())
    with open(temp, mode='w', newline='') as csvfile:
        csv.writer(csvfile).writerows(data)
    assert read_csv(temp, require_type=None)[2] is None
    with pytest.raises(ValueError, match='not recognized as'):
        assert read_csv(temp, require_type='any')
    with pytest.raises(ValueError, match='not recognized as'):
        assert read_csv(temp, require_type='points')
    with pytest.raises(ValueError, match='not recognized as'):
        read_csv(temp, require_type='shapes')


def test_csv_to_layer_data_raises(tmp_path):
    """Test various exception raising circumstances with csv_to_layer_data."""
    temp = tmp_path / 'points.csv'

    # test that points data is detected with require_type == points, any, None
    # but raises for other shape types.
    data = [['index', 'axis-0', 'axis-1']]
    data.extend(np.random.random((3, 3)).tolist())
    with open(temp, mode='w', newline='') as csvfile:
        csv.writer(csvfile).writerows(data)
    assert csv_to_layer_data(temp, require_type=None)[2] == 'points'
    assert csv_to_layer_data(temp, require_type='any')[2] == 'points'
    assert csv_to_layer_data(temp, require_type='points')[2] == 'points'
    with pytest.raises(ValueError, match='not recognized as'):
        csv_to_layer_data(temp, require_type='shapes')

    # test that unrecognized data simply returns None when require_type==None
    # but raises for specific shape types or require_type=="any"
    data = [['some', 'random', 'header']]
    data.extend(np.random.random((3, 3)).tolist())
    with open(temp, mode='w', newline='') as csvfile:
        csv.writer(csvfile).writerows(data)
    assert csv_to_layer_data(temp, require_type=None) is None
    with pytest.raises(ValueError, match='not recognized as'):
        assert csv_to_layer_data(temp, require_type='any')
    with pytest.raises(ValueError, match='not recognized as'):
        assert csv_to_layer_data(temp, require_type='points')
    with pytest.raises(ValueError, match='not recognized as'):
        csv_to_layer_data(temp, require_type='shapes')


@pytest.mark.parametrize('spec', [PNG, PNG_RGB, TIFF_3D, TIFF_2D])
@pytest.mark.parametrize('stacks', [1, 3])
def test_single_file(spec: ImageSpec, write_spec, stacks: int):
    fnames = [str(write_spec(spec)) for _ in range(stacks)]
    [(layer_data,)] = npe2.read(fnames, stack=stacks > 1)
    assert isinstance(layer_data, np.ndarray if stacks == 1 else da.Array)
    assert layer_data.shape == tuple(i for i in (stacks, *spec.shape) if i > 1)
    assert layer_data.dtype == spec.dtype


@pytest.mark.parametrize(
    'spec', [PNG, [PNG], [PNG, PNG], TIFF_3D, [TIFF_3D, TIFF_3D]]
)
@pytest.mark.parametrize('stack', [True, False])
@pytest.mark.parametrize('use_dask', [True, False, None])
def test_magic_imread(write_spec, spec: ImageSpec, stack, use_dask):
    fnames = (
        [write_spec(s) for s in spec]
        if isinstance(spec, list)
        else write_spec(spec)
    )
    images = magic_imread(fnames, stack=stack, use_dask=use_dask)
    if isinstance(spec, ImageSpec):
        expect_shape = spec.shape
    else:
        expect_shape = (len(spec), *spec[0].shape) if stack else spec[0].shape
    expect_shape = tuple(i for i in expect_shape if i > 1)

    expected_arr_type = (
        da.Array
        if (
            use_dask
            or (use_dask is None and isinstance(spec, list) and len(spec) > 1)
        )
        else np.ndarray
    )
    if isinstance(spec, list) and len(spec) > 1 and not stack:
        assert isinstance(images, list)
        assert all(isinstance(img, expected_arr_type) for img in images)
        assert all(img.shape == expect_shape for img in images)
    else:
        assert isinstance(images, expected_arr_type)
        assert images.shape == expect_shape


@pytest.mark.parametrize('stack', [True, False])
def test_irregular_images(write_spec, stack):
    specs = [PNG, PNG_RECT]
    fnames = [str(write_spec(spec)) for spec in specs]

    # Ideally, this would work "magically" with dask and irregular images,
    # but there is no foolproof way to do this without reading in all the
    # files. We need to be able to inspect the file shape without reading
    # it in first, then we can automatically turn stacking off when shapes
    # are irregular (and create proper dask arrays)
    if stack:
        with pytest.raises(
            ValueError, match='input arrays must have the same shape'
        ):
            magic_imread(fnames, use_dask=False, stack=stack)
        return

    images = magic_imread(fnames, use_dask=False, stack=stack)
    assert isinstance(images, list)
    assert len(images) == 2
    assert all(
        img.shape == spec.shape
        for img, spec in zip(images, specs, strict=False)
    )


def test_add_zarr(write_spec):
    [out] = npe2.read([str(write_spec(ZARR1))], stack=False)
    assert out[0].shape == ZARR1.shape  # type: ignore


def test_add_zarr_1d_array_is_ignored(tmp_path):
    zarr_dir = str(tmp_path / 'data.zarr')
    # For more details: https://github.com/napari/napari/issues/1471

    z = zarr.open(store=zarr_dir, mode='w')
    z.zeros(name='1d', shape=(3,), chunks=(3,), dtype='float32')

    image_path = os.path.join(zarr_dir, '1d')
    assert npe2.read([image_path], stack=False) == [(None,)]


def test_add_many_zarr_1d_array_is_ignored(tmp_path):
    # For more details: https://github.com/napari/napari/issues/1471
    zarr_dir = str(tmp_path / 'data.zarr')

    z = zarr.open(store=zarr_dir, mode='w')

    z.zeros(name='1d', shape=(3,), chunks=(3,), dtype='float32')
    z.zeros(name='2d', shape=(3, 4), chunks=(3, 4), dtype='float32')
    z.zeros(name='3d', shape=(3, 4, 5), chunks=(3, 4, 5), dtype='float32')

    for name in z.array_keys():
        [out] = npe2.read([os.path.join(zarr_dir, name)], stack=False)
        if name.endswith('1d'):
            assert out == (None,)
        else:
            assert isinstance(out[0], da.Array), name
            assert out[0].ndim == int(name[0])



"""Qt 'Debug' menu actions.

The debug menu is for developer-focused functionality that we want to be
easy-to-use and discoverable, but is not for the average user.
"""

from app_model.types import Action, KeyCode, KeyMod, SubmenuItem
from qtpy.QtCore import QTimer
from qtpy.QtWidgets import QFileDialog

from napari._app_model.constants import MenuGroup, MenuId
from napari._qt.qt_viewer import QtViewer
from napari.utils import perf
from napari.utils.history import get_save_history, update_save_history
from napari.utils.translations import trans

# Debug submenu
DEBUG_SUBMENUS = [
    (
        MenuId.MENUBAR_DEBUG,
        SubmenuItem(
            submenu=MenuId.DEBUG_PERFORMANCE,
            title=trans._('Performance Trace'),
        ),
    ),
]


def _start_trace_dialog(qt_viewer: QtViewer) -> None:
    """Open Save As dialog to start recording a trace file."""
    dlg = QFileDialog()
    hist = get_save_history()
    dlg.setHistory(hist)
    filename, _ = dlg.getSaveFileName(
        qt_viewer,  # parent
        trans._('Record performance trace file'),  # caption
        hist[0],  # directory in PyQt, dir in PySide
        filter=trans._('Trace Files (*.json)'),
    )
    if filename:
        if not filename.endswith('.json'):
            filename += '.json'

        # Schedule this to avoid bogus "MetaCall" event for the entire
        # time the file dialog was up.
        QTimer.singleShot(0, lambda: _start_trace(filename))

        update_save_history(filename)


def _start_trace(path: str) -> None:
    """Start recording a trace file."""
    perf.timers.start_trace_file(path)


def _stop_trace() -> None:
    """Stop recording a trace file."""
    perf.timers.stop_trace_file()


def _is_set_trace_active() -> bool:
    """Whether we are currently recording a set trace."""
    return perf.timers.trace_file is not None


Q_DEBUG_ACTIONS: list[Action] = [
    Action(
        id='napari.window.debug.start_trace_dialog',
        title=trans._('Start Recording...'),
        callback=_start_trace_dialog,
        menus=[
            {'id': MenuId.DEBUG_PERFORMANCE, 'group': MenuGroup.NAVIGATION}
        ],
        keybindings=[{'primary': KeyMod.Alt | KeyCode.KeyT}],
        enablement='not is_set_trace_active',
        status_tip=trans._('Start recording a trace file'),
    ),
    Action(
        id='napari.window.debug.stop_trace',
        title=trans._('Stop Recording...'),
        callback=_stop_trace,
        menus=[
            {'id': MenuId.DEBUG_PERFORMANCE, 'group': MenuGroup.NAVIGATION}
        ],
        keybindings=[{'primary': KeyMod.Alt | KeyMod.Shift | KeyCode.KeyT}],
        enablement='is_set_trace_active',
        status_tip=trans._('Stop recording a trace file'),
    ),
]



from __future__ import annotations

import sys
import warnings
from collections.abc import Callable
from contextlib import contextmanager
from importlib.metadata import distributions
from typing import TYPE_CHECKING, NamedTuple

from napari.settings._fields import Version
from napari.settings._shortcuts import ShortcutsSettings
from napari.utils.triangulation_backend import TriangulationBackend

if TYPE_CHECKING:
    from napari.settings._napari_settings import NapariSettings

_MIGRATORS: list[Migrator] = []
MigratorF = Callable[['NapariSettings'], None]


class Migrator(NamedTuple):
    """Tuple of from-version, to-version, migrator function."""

    from_: Version
    to_: Version
    run: MigratorF


def do_migrations(model: NapariSettings):
    """Migrate (update) a NapariSettings model in place."""
    for migration in sorted(_MIGRATORS, key=lambda m: m.from_):
        if model.schema_version == migration.from_:
            with mutation_allowed(model):
                backup = model.dict()
                try:
                    migration.run(model)
                    model.schema_version = migration.to_
                except Exception as e:  # noqa BLE001
                    msg = (
                        f'Failed to migrate settings from v{migration.from_} '
                        f'to v{migration.to_}. Error: {e}. '
                    )
                    try:
                        model.update(backup)
                        msg += 'You may need to reset your settings with `napari --reset`. '
                    except Exception:  # noqa BLE001
                        msg += 'Settings rollback also failed. Please run `napari --reset`.'
                    warnings.warn(msg)
                    return
    model._maybe_save()


@contextmanager
def mutation_allowed(obj: NapariSettings):
    """Temporarily allow mutations on an immutable model."""
    config = obj.__config__
    prev, config.allow_mutation = config.allow_mutation, True
    try:
        yield
    finally:
        config.allow_mutation = prev


def migrator(from_: str, to_: str) -> Callable[[MigratorF], MigratorF]:
    """Decorate function as migrating settings from v `from_` to v `to_`.

    A migrator should mutate a `NapariSettings` model from schema version
    `from_` to schema version `to_` (in place).

    Parameters
    ----------
    from_ : str
        NapariSettings.schema_version version that this migrator expects as
        input
    to_ : str
        NapariSettings.schema_version version after this migrator has been
        executed.

    Returns
    -------
    Callable[ [MigratorF], MigratorF ]
        _description_
    """

    def decorator(migrate_func: MigratorF) -> MigratorF:
        _from, _to = Version.parse(from_), Version.parse(to_)
        assert _to >= _from, 'Migrator must increase the version.'
        _MIGRATORS.append(Migrator(_from, _to, migrate_func))
        return migrate_func

    return decorator


@migrator('0.3.0', '0.4.0')
def v030_v040(model: NapariSettings):
    """Migrate from v0.3.0 to v0.4.0.

    Prior to v0.4.0, npe2 plugins were automatically added to disabled plugins.
    This migration removes any npe2 plugins discovered in the environment
    (at migration time) from the "disabled plugins" set.
    """
    for dist in distributions():
        for ep in dist.entry_points:
            if ep.group == 'napari.manifest':
                model.plugins.disabled_plugins.discard(dist.metadata['Name'])


@migrator('0.4.0', '0.5.0')
def v040_050(model: NapariSettings):
    """Migrate from v0.4.0 to v0.5.0

    Prior to 0.5.0 existing preferences may have reader extensions
    preferences saved without a leading *.
    fnmatch would fail on these so we coerce them to include a *
    e.g. '.csv' becomes '*.csv'
    """
    from napari.settings._utils import _coerce_extensions_to_globs

    current_settings = model.plugins.extension2reader
    new_settings = _coerce_extensions_to_globs(current_settings)
    model.plugins.extension2reader = new_settings


def _swap_ctrl_cmd(keybinding):
    """Swap the Control and Command/Super/Meta modifiers in a keybinding.

    See `v050_060` for motivation.
    """
    from napari.utils.key_bindings import KeyBinding

    kb = KeyBinding.from_str(
        str(keybinding)
        .replace('Ctrl', 'Temp')
        .replace('Meta', 'Ctrl')
        .replace('Temp', 'Meta')
    )
    return kb


@migrator('0.5.0', '0.6.0')
def v050_060(model: NapariSettings):
    """Migrate from v0.5.0 to v0.6.0.

    In #5103 we went from using our own keybinding model to using app-model's.
    Several consequences of this are:
    - Control is written out as Ctrl (and that is the only valid input)
    - Option is written out as Alt (and that is the only valid input)
    - Super/Command/Cmd are written out as Meta (both Meta and Cmd are valid
      inputs)
    - modifiers with keys are written as Mod+Key rather than Mod-Key. However,
      both versions are valid inputs.
    - macOS shortcuts using command keys are written out as Meta, where they
      used to be written out as Control.

    The alias problem is solved in `napari.utils.keybindings.coerce_keybinding`
    by substituting all variants with the canonical versions in app-model.

    The separator problem (-/+) can be ignored.

    This migrator solves the final problem, by detecting whether the current
    OS is macOS, and swapping Control and Meta in all key bindings if so.
    """
    if sys.platform == 'darwin':
        current_keybinds = model.shortcuts.shortcuts
        default_shortcuts = ShortcutsSettings().shortcuts
        new_keybinds = {}
        for action_str, keybind_list in current_keybinds.items():
            new_keybind_list = []
            for kb in keybind_list:
                if kb not in default_shortcuts[action_str]:
                    new_keybind_list.append(_swap_ctrl_cmd(kb))
                else:
                    new_keybind_list.append(kb)
            new_keybinds[action_str] = new_keybind_list
        model.shortcuts.shortcuts = new_keybinds


@migrator('0.6.0', '0.7.0')
def v060_070(model: NapariSettings):
    """Migrate from v0.6.0 to v0.7.0.

    In #7627 we updated the default value of the npe2 shim, which bumped the
    minor version. However, the migration doesn't need to do anything about
    that.

    In #7747, we changed from `experimental.compiled_triangulation` (bool) to
    `experimental.triangulation_backend` (enum). We need to migrate the bool
    setting to oen of the backends.

    Note: both PRs occurred between 0.5.6 and 0.6.0, so they require only a
    single version bump/migration.
    """
    model.experimental.triangulation_backend = (
        TriangulationBackend.partsegcore
        if model.experimental.compiled_triangulation
        else TriangulationBackend.fastest_available  # must match default setting
    )
    model.experimental.compiled_triangulation = False  # reset to default



"""
These tests cover and help explain the implementations of different
types of generic encodings, like constant, manual, and derived encodings,
rather than the types of values they encode like strings and colors
or the ways those are encoded.

In particular, these cover the stateful part of the StyleEncoding, which
is important to napari at the time of writing, but may be removed in the future.
"""

from typing import Any

import numpy as np
import pandas as pd
import pytest

from napari._pydantic_compat import Field
from napari.layers.utils.style_encoding import (
    _ConstantStyleEncoding,
    _DerivedStyleEncoding,
    _ManualStyleEncoding,
)
from napari.utils.events.custom_types import Array


@pytest.fixture
def features() -> pd.DataFrame:
    return pd.DataFrame(
        {
            'scalar': [1, 2, 3],
            'vector': [[1, 1], [2, 2], [3, 3]],
        }
    )


Scalar = Array[int, ()]
ScalarArray = Array[int, (-1,)]


class ScalarConstantEncoding(_ConstantStyleEncoding[Scalar, ScalarArray]):
    constant: Scalar


def test_scalar_constant_encoding_apply(features):
    encoding = ScalarConstantEncoding(constant=0)
    encoding._apply(features)
    np.testing.assert_array_equal(encoding._values, 0)


def test_scalar_constant_encoding_append():
    encoding = ScalarConstantEncoding(constant=0)
    encoding._append(Vector.validate_type([4, 5]))
    np.testing.assert_array_equal(encoding._values, 0)


def test_scalar_constant_encoding_delete():
    encoding = ScalarConstantEncoding(constant=0)
    encoding._delete([0, 2])
    np.testing.assert_array_equal(encoding._values, 0)


def test_scalar_constant_encoding_clear():
    encoding = ScalarConstantEncoding(constant=0)
    encoding._clear()
    np.testing.assert_array_equal(encoding._values, 0)


class ScalarManualEncoding(_ManualStyleEncoding[Scalar, ScalarArray]):
    array: ScalarArray
    default: Scalar = np.array(-1)


def test_scalar_manual_encoding_apply_with_shorter(features):
    encoding = ScalarManualEncoding(array=[1, 2, 3, 4])
    encoding._apply(features)
    np.testing.assert_array_equal(encoding._values, [1, 2, 3])


def test_scalar_manual_encoding_apply_with_equal_length(features):
    encoding = ScalarManualEncoding(array=[1, 2, 3])
    encoding._apply(features)
    np.testing.assert_array_equal(encoding._values, [1, 2, 3])


def test_scalar_manual_encoding_apply_with_longer(features):
    encoding = ScalarManualEncoding(array=[1, 2], default=-1)
    encoding._apply(features)
    np.testing.assert_array_equal(encoding._values, [1, 2, -1])


def test_scalar_manual_encoding_append():
    encoding = ScalarManualEncoding(array=[1, 2, 3])
    encoding._append(Vector.validate_type([4, 5]))
    np.testing.assert_array_equal(encoding._values, [1, 2, 3, 4, 5])


def test_scalar_manual_encoding_delete():
    encoding = ScalarManualEncoding(array=[1, 2, 3])
    encoding._delete([0, 2])
    np.testing.assert_array_equal(encoding._values, [2])


def test_scalar_manual_encoding_clear():
    encoding = ScalarManualEncoding(array=[1, 2, 3])
    encoding._clear()
    np.testing.assert_array_equal(encoding._values, [1, 2, 3])


class ScalarDirectEncoding(_DerivedStyleEncoding[Scalar, ScalarArray]):
    feature: str
    fallback: Scalar = np.array(-1)

    def __call__(self, features: Any) -> ScalarArray:
        return ScalarArray.validate_type(features[self.feature])


def test_scalar_derived_encoding_apply(features):
    encoding = ScalarDirectEncoding(feature='scalar')

    encoding._apply(features)

    expected_values = features['scalar']
    np.testing.assert_array_equal(encoding._values, expected_values)


def test_scalar_derived_encoding_apply_with_failure(features):
    encoding = ScalarDirectEncoding(feature='not_a_column', fallback=-1)

    with pytest.warns(RuntimeWarning):
        encoding._apply(features)

    np.testing.assert_array_equal(encoding._values, [-1] * len(features))


def test_scalar_derived_encoding_append():
    encoding = ScalarDirectEncoding(feature='scalar')
    encoding._cached = ScalarArray.validate_type([1, 2, 3])

    encoding._append(ScalarArray.validate_type([4, 5]))

    np.testing.assert_array_equal(encoding._values, [1, 2, 3, 4, 5])


def test_scalar_derived_encoding_delete():
    encoding = ScalarDirectEncoding(feature='scalar')
    encoding._cached = ScalarArray.validate_type([1, 2, 3])

    encoding._delete([0, 2])

    np.testing.assert_array_equal(encoding._values, [2])


def test_scalar_derived_encoding_clear():
    encoding = ScalarDirectEncoding(feature='scalar')
    encoding._cached = ScalarArray.validate_type([1, 2, 3])

    encoding._clear()

    np.testing.assert_array_equal(encoding._values, [])


Vector = Array[int, (2,)]
VectorArray = Array[int, (-1, 2)]


class VectorConstantEncoding(_ConstantStyleEncoding[Vector, VectorArray]):
    constant: Vector


def test_vector_constant_encoding_apply(features):
    encoding = VectorConstantEncoding(constant=[0, 0])
    encoding._apply(features)
    np.testing.assert_array_equal(encoding._values, [0, 0])


def test_vector_constant_encoding_append():
    encoding = VectorConstantEncoding(constant=[0, 0])
    encoding._append(Vector.validate_type([4, 5]))
    np.testing.assert_array_equal(encoding._values, [0, 0])


def test_vector_constant_encoding_delete():
    encoding = VectorConstantEncoding(constant=[0, 0])
    encoding._delete([0, 2])
    np.testing.assert_array_equal(encoding._values, [0, 0])


def test_vector_constant_encoding_clear():
    encoding = VectorConstantEncoding(constant=[0, 0])
    encoding._clear()
    np.testing.assert_array_equal(encoding._values, [0, 0])


class VectorManualEncoding(_ManualStyleEncoding[Vector, VectorArray]):
    array: VectorArray
    default: Vector = Field(default_factory=lambda: np.array([-1, -1]))


def test_vector_manual_encoding_apply_with_shorter(features):
    encoding = VectorManualEncoding(array=[[1, 1], [2, 2], [3, 3], [4, 4]])
    encoding._apply(features)
    np.testing.assert_array_equal(encoding._values, [[1, 1], [2, 2], [3, 3]])


def test_vector_manual_encoding_apply_with_equal_length(features):
    encoding = VectorManualEncoding(array=[[1, 1], [2, 2], [3, 3]])
    encoding._apply(features)
    np.testing.assert_array_equal(encoding._values, [[1, 1], [2, 2], [3, 3]])


def test_vector_manual_encoding_apply_with_longer(features):
    encoding = VectorManualEncoding(array=[[1, 1], [2, 2]], default=[-1, -1])
    encoding._apply(features)
    np.testing.assert_array_equal(encoding._values, [[1, 1], [2, 2], [-1, -1]])


def test_vector_manual_encoding_append():
    encoding = VectorManualEncoding(array=[[1, 1], [2, 2], [3, 3]])
    encoding._append(Vector.validate_type([[4, 4], [5, 5]]))
    np.testing.assert_array_equal(
        encoding._values, [[1, 1], [2, 2], [3, 3], [4, 4], [5, 5]]
    )


def test_vector_manual_encoding_delete():
    encoding = VectorManualEncoding(array=[[1, 1], [2, 2], [3, 3]])
    encoding._delete([0, 2])
    np.testing.assert_array_equal(encoding._values, [[2, 2]])


def test_vector_manual_encoding_clear():
    encoding = VectorManualEncoding(array=[[1, 1], [2, 2], [3, 3]])
    encoding._clear()
    np.testing.assert_array_equal(encoding._values, [[1, 1], [2, 2], [3, 3]])


class VectorDirectEncoding(_DerivedStyleEncoding[Vector, VectorArray]):
    feature: str
    fallback: Vector = Field(default_factory=lambda: np.array([-1, -1]))

    def __call__(self, features: Any) -> Vector | VectorArray:
        return VectorArray.validate_type(list(features[self.feature]))


def test_vector_derived_encoding_apply(features):
    encoding = VectorDirectEncoding(feature='vector')

    encoding._apply(features)

    expected_values = list(features['vector'])
    np.testing.assert_array_equal(encoding._values, expected_values)


def test_vector_derived_encoding_apply_with_failure(features):
    encoding = VectorDirectEncoding(feature='not_a_column', fallback=[-1, -1])

    with pytest.warns(RuntimeWarning):
        encoding._apply(features)

    np.testing.assert_array_equal(encoding._values, [[-1, -1]] * len(features))


def test_vector_derived_encoding_append():
    encoding = VectorDirectEncoding(feature='vector')
    encoding._cached = VectorArray.validate_type([[1, 1], [2, 2], [3, 3]])

    encoding._append(VectorArray.validate_type([[4, 4], [5, 5]]))

    np.testing.assert_array_equal(
        encoding._values, [[1, 1], [2, 2], [3, 3], [4, 4], [5, 5]]
    )


def test_vector_derived_encoding_delete():
    encoding = VectorDirectEncoding(feature='vector')
    encoding._cached = VectorArray.validate_type([[1, 1], [2, 2], [3, 3]])

    encoding._delete([0, 2])

    np.testing.assert_array_equal(encoding._values, [[2, 2]])


def test_vector_derived_encoding_clear():
    encoding = VectorDirectEncoding(feature='vector')
    encoding._cached = VectorArray.validate_type([[1, 1], [2, 2], [3, 3]])

    encoding._clear()

    np.testing.assert_array_equal(encoding._values, np.empty((0, 2)))



"""Contains napari color constants and utilities."""

from collections.abc import Callable, Iterator
from typing import Union

import numpy as np

from napari.utils.colormaps.standardize_color import transform_color

ColorValueParam = Union[np.ndarray, list, tuple, str, None]
ColorArrayParam = Union[np.ndarray, list, tuple, None]


class ColorValue(np.ndarray):
    """A custom pydantic field type for storing one color value.

    Using this as a field type in a pydantic model means that validation
    of that field (e.g. on initialization or setting) will automatically
    use the ``validate`` method to coerce a value to a single color.
    """

    def __new__(cls, value: ColorValueParam) -> 'ColorValue':
        return cls.validate(value)

    @classmethod
    def __get_validators__(
        cls,
    ) -> Iterator[Callable[[ColorValueParam], 'ColorValue']]:
        yield cls.validate

    @classmethod
    def validate(cls, value: ColorValueParam) -> 'ColorValue':
        """Validates and coerces the given value into an array storing one color.

        Parameters
        ----------
        value : Union[np.ndarray, list, tuple, str, None]
            A supported single color value, which must be one of the following.

            - A supported RGB(A) sequence of floating point values in [0, 1].
            - A CSS3 color name: https://www.w3.org/TR/css-color-3/#svg-color
            - A single character matplotlib color name: https://matplotlib.org/stable/tutorials/colors/colors.html#specifying-colors
            - An RGB(A) hex code string.

        Returns
        -------
        np.ndarray
            An RGBA color vector of floating point values in [0, 1].

        Raises
        ------
        ValueError, AttributeError, KeyError
            If the value is not recognized as a color.

        Examples
        --------
        Coerce an RGBA array-like.

        >>> ColorValue.validate([1, 0, 0, 1])
        array([1., 0., 0., 1.], dtype=float32)

        Coerce a CSS3 color name.

        >>> ColorValue.validate('red')
        array([1., 0., 0., 1.], dtype=float32)

        Coerce a matplotlib single character color name.

        >>> ColorValue.validate('r')
        array([1., 0., 0., 1.], dtype=float32)

        Coerce an RGB hex-code.

        >>> ColorValue.validate('#ff0000')
        array([1., 0., 0., 1.], dtype=float32)
        """
        return transform_color(value)[0].view(cls)


class ColorArray(np.ndarray):
    """A custom pydantic field type for storing an array of color values.

    Using this as a field type in a pydantic model means that validation
    of that field (e.g. on initialization or setting) will automatically
    use the ``validate`` method to coerce a value to an array of colors.
    """

    def __new__(cls, value: ColorArrayParam) -> 'ColorArray':
        return cls.validate(value)

    @classmethod
    def __get_validators__(
        cls,
    ) -> Iterator[Callable[[ColorArrayParam], 'ColorArray']]:
        yield cls.validate

    def __sizeof__(self) -> int:
        return super().__sizeof__() + self.nbytes

    @classmethod
    def validate(cls, value: ColorArrayParam) -> 'ColorArray':
        """Validates and coerces the given value into an array storing many colors.

        Parameters
        ----------
        value : Union[np.ndarray, list, tuple, None]
            A supported sequence of single color values.
            See ``ColorValue.validate`` for valid single color values.
            In general each value should be of the same type, so avoid
            passing values like ``['red', [0, 0, 1]]``.

        Returns
        -------
        np.ndarray
            An array of N colors where each row is an RGBA color vector with
            floating point values in [0, 1].

        Raises
        ------
        ValueError, AttributeError, KeyError
            If the value is not recognized as an array of colors.

        Examples
        --------
        Coerce a list of CSS3 color names.

        >>> ColorArray.validate(['red', 'blue'])
        array([[1., 0., 0., 1.],
               [0., 0., 1., 1.]], dtype=float32)

        Coerce a tuple of matplotlib single character color names.

        >>> ColorArray.validate(('r', 'b'))
        array([[1., 0., 0., 1.],
               [0., 0., 1., 1.]], dtype=float32)
        """
        # Special case an empty supported sequence because transform_color
        # warns and returns an array containing a default color in that case.
        if isinstance(value, np.ndarray | list | tuple) and len(value) == 0:
            return np.empty((0, 4), np.float32).view(cls)
        return transform_color(value).view(cls)



from __future__ import annotations

from typing import Any

import numpy as np
from vispy.scene.widgets.viewbox import ViewBox

from napari._vispy.layers.base import VispyBaseLayer
from napari._vispy.layers.image import VispyImageLayer
from napari._vispy.layers.labels import VispyLabelsLayer
from napari._vispy.layers.points import VispyPointsLayer
from napari._vispy.layers.shapes import VispyShapesLayer
from napari._vispy.layers.surface import VispySurfaceLayer
from napari._vispy.layers.tracks import VispyTracksLayer
from napari._vispy.layers.vectors import VispyVectorsLayer
from napari._vispy.overlays.axes import VispyAxesOverlay
from napari._vispy.overlays.base import VispyBaseOverlay
from napari._vispy.overlays.bounding_box import VispyBoundingBoxOverlay
from napari._vispy.overlays.brush_circle import VispyBrushCircleOverlay
from napari._vispy.overlays.interaction_box import (
    VispySelectionBoxOverlay,
    VispyTransformBoxOverlay,
)
from napari._vispy.overlays.labels_polygon import VispyLabelsPolygonOverlay
from napari._vispy.overlays.scale_bar import VispyScaleBarOverlay
from napari._vispy.overlays.text import VispyTextOverlay
from napari.components.overlays import (
    AxesOverlay,
    BoundingBoxOverlay,
    BrushCircleOverlay,
    LabelsPolygonOverlay,
    Overlay,
    ScaleBarOverlay,
    SelectionBoxOverlay,
    TextOverlay,
    TransformBoxOverlay,
)
from napari.layers import (
    Image,
    Labels,
    Layer,
    Points,
    Shapes,
    Surface,
    Tracks,
    Vectors,
)
from napari.utils.translations import trans

layer_to_visual: dict[type[Layer], type[VispyBaseLayer]] = {
    Image: VispyImageLayer,
    Labels: VispyLabelsLayer,
    Points: VispyPointsLayer,
    Shapes: VispyShapesLayer,
    Surface: VispySurfaceLayer,
    Vectors: VispyVectorsLayer,
    Tracks: VispyTracksLayer,
}


overlay_to_visual: dict[type[Overlay], type[VispyBaseOverlay]] = {
    ScaleBarOverlay: VispyScaleBarOverlay,
    TextOverlay: VispyTextOverlay,
    AxesOverlay: VispyAxesOverlay,
    BoundingBoxOverlay: VispyBoundingBoxOverlay,
    TransformBoxOverlay: VispyTransformBoxOverlay,
    SelectionBoxOverlay: VispySelectionBoxOverlay,
    BrushCircleOverlay: VispyBrushCircleOverlay,
    LabelsPolygonOverlay: VispyLabelsPolygonOverlay,
}


def create_vispy_layer(
    layer: Layer, *args: Any, **kwargs: Any
) -> VispyBaseLayer:
    """Create vispy visual for a layer based on its layer type.

    Parameters
    ----------
    layer : napari.layers._base_layer.Layer
        Layer that needs its property widget created.

    Returns
    -------
    visual : VispyBaseLayer
        Vispy layer
    """
    # find the closest parent class, to maintain behaviour from #2757
    for cls in layer.__class__.mro():
        if cls in layer_to_visual:
            return layer_to_visual[cls](layer, *args, **kwargs)

    raise TypeError(
        trans._(
            'Could not find VispyLayer for layer of type {dtype}',
            deferred=True,
            dtype=type(layer),
        )
    )


def create_vispy_overlay(overlay: Overlay, **kwargs) -> VispyBaseOverlay:
    """
    Create vispy visual for Overlay based on its type.

    Parameters
    ----------
    overlay : napari.components.overlays.VispyBaseOverlay
        The overlay to create a visual for.

    Returns
    -------
    visual : VispyBaseOverlay
        Vispy overlay
    """
    for cls in overlay.__class__.mro():
        if cls in overlay_to_visual:
            return overlay_to_visual[cls](overlay=overlay, **kwargs)

    raise TypeError(
        trans._(
            'Could not find VispyOverlay for overlay of type {dtype}',
            deferred=True,
            dtype=type(overlay),
        )
    )


def get_view_direction_in_scene_coordinates(
    view: ViewBox,
    ndim: int,
    dims_displayed: tuple[int],
) -> np.ndarray | None:
    """Calculate the unit vector pointing in the direction of the view.

    This is only for 3D viewing, so it returns None when
    len(dims_displayed) == 2.
    Adapted From:
    https://stackoverflow.com/questions/37877592/
        get-view-direction-relative-to-scene-in-vispy/37882984

    Parameters
    ----------
    view : vispy.scene.widgets.viewbox.ViewBox
        The vispy view box object to get the view direction from.
    ndim : int
        The number of dimensions in the full nD dims model.
        This is typically from viewer.dims.ndim
    dims_displayed : Tuple[int]
        The indices of the dims displayed in the viewer.
        This is typically from viewer.dims.displayed.

    Returns
    -------
    view_vector : np.ndarray
        Unit vector in the direction of the view in scene coordinates.
        Axes are ordered zyx. If the viewer is in 2D
        (i.e., len(dims_displayed) == 2), view_vector is None.
    """
    # only return a vector when viewing in 3D
    if len(dims_displayed) == 2:
        return None

    tform = view.scene.transform
    w, h = view.canvas.size

    # get a point at the center of the canvas
    # (homogeneous screen coords)
    screen_center = np.array([w / 2, h / 2, 0, 1])

    # find a point just in front of the center point
    # transform both to world coords and find the vector
    d1 = np.array([0, 0, 1, 0])
    point_in_front_of_screen_center = screen_center + d1
    p1 = tform.imap(point_in_front_of_screen_center)
    p0 = tform.imap(screen_center)
    d2 = p1 - p0

    # in 3D world coordinates
    d3 = d2[:3]
    d4 = d3 / np.linalg.norm(d3)

    # data are ordered xyz on vispy Volume
    d4 = d4[[2, 1, 0]]
    view_dir_world = np.zeros((ndim,))
    for i, d in enumerate(dims_displayed):
        view_dir_world[d] = d4[i]

    return view_dir_world



import html
from unittest.mock import Mock

import numpy as np
import pytest

from napari._tests.utils import skip_on_win_ci
from napari._version import __version__
from napari.utils import nbscreenshot


@skip_on_win_ci
def test_nbscreenshot(make_napari_viewer):
    """Test taking a screenshot."""
    viewer = make_napari_viewer()

    np.random.seed(0)
    data = np.random.random((10, 15))
    viewer.add_image(data)

    rich_display_object = nbscreenshot(viewer)
    assert hasattr(rich_display_object, '_repr_png_')
    # Trigger method that would run in jupyter notebook cell automatically
    png_bytes = rich_display_object._repr_png_()
    assert rich_display_object.image is not None
    # Test digital watermark is included in bytes of .png file
    version_byte_string = __version__.encode('utf-8')
    assert b'napari version' in png_bytes
    assert version_byte_string in png_bytes


@skip_on_win_ci
@pytest.mark.parametrize(
    ('alt_text_input', 'expected_alt_text'),
    [
        (None, None),
        ('Good alt text', 'Good alt text'),
        # Naughty strings https://github.com/minimaxir/big-list-of-naughty-strings
        # ASCII punctuation
        (r",./;'[]\-=", ',./;&#x27;[]\\-='),
        # ASCII punctuation 2, skipping < because that is interpreted as the start
        # of an HTML element.
        ('>?:"{}|_+', '&gt;?:&quot;{}|_+'),
        ('!@#$%^&*()`~', '!@#$%^&amp;*()`~'),  # ASCII punctuation 3
        # # Emojis
        ('😍', '😍'),  # emoji 1
        (
            '👨‍🦰 👨🏿‍🦰 👨‍🦱 👨🏿‍🦱 🦹🏿‍♂️',
            '👨‍🦰 👨🏿‍🦰 👨‍🦱 👨🏿‍🦱 🦹🏿‍♂️',
        ),  # emoji 2
        (r'¯\_(ツ)_/¯', '¯\\_(ツ)_/¯'),  # Japanese emoticon
        # # Special characters
        (
            '田中さんにあげて下さい',
            '田中さんにあげて下さい',
        ),  # two-byte characters
        (
            '表ポあA鷗ŒéＢ逍Üßªąñ丂㐀𠀀',  # noqa: RUF001
            '表ポあA鷗ŒéＢ逍Üßªąñ丂㐀𠀀',  # noqa: RUF001
        ),  # special unicode chars
        ('گچپژ', 'گچپژ'),  # Persian special characters
        # # Script injection
        ('<script>alert(0)</script>', None),  # script injection 1
        ('&lt;script&gt;alert(&#39;1&#39;);&lt;/script&gt;', None),
        ('<svg><script>123<1>alert(3)</script>', None),
    ],
)
def test_safe_alt_text(alt_text_input, expected_alt_text):
    display_obj = nbscreenshot(Mock(), alt_text=alt_text_input)
    if not expected_alt_text:
        assert not display_obj.alt_text
    else:
        assert html.escape(display_obj.alt_text) == expected_alt_text


def test_invalid_alt_text():
    with pytest.warns(
        UserWarning,
        match='The provided alt text does not constitute valid html',
    ):
        # because string with only whitespace messes up with the parser
        display_obj = nbscreenshot(Mock(), alt_text=' ')
    assert display_obj.alt_text is None

    with pytest.warns(
        UserWarning,
        match='The provided alt text does not constitute valid html',
    ):
        # because string with only whitespace messes up with the parser
        display_obj = nbscreenshot(Mock(), alt_text='')
    assert display_obj.alt_text is None



from typing import Union, cast

from napari._pydantic_compat import Field
from napari.settings._fields import Theme
from napari.utils.events.evented_model import ComparisonDelayer, EventedModel
from napari.utils.theme import available_themes, get_theme
from napari.utils.translations import trans


class HighlightSettings(EventedModel):
    highlight_thickness: int = Field(
        1,
        title=trans._('Highlight thickness'),
        description=trans._(
            'Select the highlight thickness when hovering over shapes/points.'
        ),
        ge=1,
        le=10,
    )
    highlight_color: list[float] = Field(
        [0.0, 0.6, 1.0, 1.0],
        title=trans._('Highlight color'),
        description=trans._(
            'Select the highlight color when hovering over shapes/points.'
        ),
    )


class AppearanceSettings(EventedModel):
    theme: Theme = Field(
        Theme('dark'),
        title=trans._('Theme'),
        description=trans._('Select the user interface theme.'),
        env='napari_theme',
    )
    font_size: int = Field(
        int(get_theme('dark').font_size[:-2]),
        title=trans._('Font size'),
        description=trans._('Select the user interface font size.'),
        ge=5,
        le=20,
    )
    highlight: HighlightSettings = Field(
        HighlightSettings(),
        title=trans._('Highlight'),
        description=trans._(
            'Select the highlight color and thickness to use when hovering over shapes/points.'
        ),
    )
    layer_tooltip_visibility: bool = Field(
        False,
        title=trans._('Show layer tooltips'),
        description=trans._('Toggle to display a tooltip on mouse hover.'),
    )
    update_status_based_on_layer: bool = Field(
        True,
        title=trans._('Update status based on layer'),
        description=trans._(
            'Calculate status bar based on current active layer and mouse position.'
        ),
    )

    def update(
        self, values: Union['EventedModel', dict], recurse: bool = True
    ) -> None:
        if isinstance(values, self.__class__):
            values = values.dict()
        values = cast(dict, values)

        # Check if a font_size change is needed when changing theme:
        # If the font_size setting doesn't correspond to the default value
        # of the current theme no change is done, otherwise
        # the font_size value is set to the new selected theme font size value
        if 'theme' in values and values['theme'] != self.theme:
            current_theme = get_theme(self.theme)
            new_theme = get_theme(values['theme'])
            if values['font_size'] == int(current_theme.font_size[:-2]):
                values['font_size'] = int(new_theme.font_size[:-2])
        super().update(values, recurse)

    def __setattr__(self, key: str, value: Theme) -> None:
        # Check if a font_size change is needed when changing theme:
        # If the font_size setting doesn't correspond to the default value
        # of the current theme no change is done, otherwise
        # the font_size value is set to the new selected theme font size value
        if key == 'theme' and value != self.theme:
            with ComparisonDelayer(self):
                new_theme = None
                current_theme = None
                if value in available_themes():
                    new_theme = get_theme(value)
                if self.theme in available_themes():
                    current_theme = get_theme(self.theme)
                if (
                    new_theme
                    and current_theme
                    and self.font_size == int(current_theme.font_size[:-2])
                ):
                    self.font_size = int(new_theme.font_size[:-2])
                super().__setattr__(key, value)
        else:
            super().__setattr__(key, value)

    class NapariConfig:
        # Napari specific configuration
        preferences_exclude = ('schema_version',)

    def refresh_themes(self) -> None:
        """Updates theme data.
        This is not a fantastic solution but it works. Each time a new theme is
        added (either by a plugin or directly by the user) the enum is updated in
        place, ensuring that Preferences dialog can still be opened.
        """
        self.schema()['properties']['theme'].update(enum=available_themes())



import copy
import itertools
import time
from collections import defaultdict
from dataclasses import dataclass
from importlib.metadata import version

import numpy as np
import numpy.testing as npt
import pandas as pd
import pytest
import xarray as xr
import zarr
from packaging.version import parse as parse_version
from skimage import data as sk_data

from napari._tests.utils import check_layer_world_data_extent
from napari.components import ViewerModel
from napari.components.dims import Dims
from napari.layers import Labels
from napari.layers.labels._labels_constants import LabelsRendering
from napari.layers.labels._labels_utils import get_contours
from napari.utils import Colormap
from napari.utils._test_utils import (
    validate_all_params_in_docstring,
    validate_kwargs_sorted,
)
from napari.utils.colormaps import (
    CyclicLabelColormap,
    DirectLabelColormap,
    label_colormap,
)


@pytest.fixture
def direct_colormap():
    """Return a DirectLabelColormap."""
    return DirectLabelColormap(
        color_dict={
            0: [0, 0, 0, 0],
            1: [1, 0, 0, 1],
            2: [0, 1, 0, 1],
            None: [0, 0, 1, 1],
        }
    )


@pytest.fixture
def random_colormap():
    """Return a LabelColormap."""
    return label_colormap(50)


def test_random_labels():
    """Test instantiating Labels layer with random 2D data."""
    shape = (10, 15)
    np.random.seed(0)
    data = np.random.randint(20, size=shape)
    layer = Labels(data)
    np.testing.assert_array_equal(layer.data, data)
    assert layer.ndim == len(shape)
    np.testing.assert_array_equal(layer.extent.data[1], [s - 1 for s in shape])
    assert layer._data_view.shape == shape[-2:]
    assert layer.editable is True


def test_all_zeros_labels():
    """Test instantiating Labels layer with all zeros data."""
    shape = (10, 15)
    data = np.zeros(shape, dtype=int)
    layer = Labels(data)
    np.testing.assert_array_equal(layer.data, data)
    assert layer.ndim == len(shape)
    np.testing.assert_array_equal(layer.extent.data[1], [s - 1 for s in shape])
    assert layer._data_view.shape == shape[-2:]


def test_3D_labels():
    """Test instantiating Labels layer with random 3D data."""
    shape = (6, 10, 15)
    np.random.seed(0)
    data = np.random.randint(20, size=shape)
    layer = Labels(data)
    np.testing.assert_array_equal(layer.data, data)
    assert layer.ndim == len(shape)
    np.testing.assert_array_equal(layer.extent.data[1], [s - 1 for s in shape])
    assert layer._data_view.shape == shape[-2:]
    assert layer.editable is True

    layer._slice_dims(Dims(ndim=3, ndisplay=3))
    assert layer.editable is True
    assert layer.mode == 'pan_zoom'


def test_float_labels():
    """Test instantiating labels layer with floats"""
    np.random.seed(0)
    data = np.random.uniform(0, 20, size=(10, 10))
    with pytest.raises(TypeError):
        Labels(data)

    data0 = np.random.uniform(20, size=(20, 20))
    data1 = data0[::2, ::2].astype(np.int32)
    data = [data0, data1]
    with pytest.raises(TypeError):
        Labels(data)


def test_bool_labels():
    """Test instantiating labels layer with bools"""
    data = np.zeros((10, 10), dtype=bool)
    layer = Labels(data)
    assert np.issubdtype(layer.data.dtype, np.integer)

    data0 = np.zeros((20, 20), dtype=bool)
    data1 = data0[::2, ::2].astype(np.int32)
    data = [data0, data1]
    layer = Labels(data)
    assert all(np.issubdtype(d.dtype, np.integer) for d in layer.data)


def test_editing_bool_labels():
    # make random data, mostly 0s
    data = np.random.random((10, 10)) > 0.7
    # create layer, which may convert bool to uint8 *as a view*
    layer = Labels(data)
    # paint the whole layer with 1
    layer.paint_polygon(
        points=[[-1, -1], [-1, 11], [11, 11], [11, -1]],
        new_label=1,
    )
    # check that the original data has been correspondingly modified
    assert np.all(data)


def test_changing_labels():
    """Test changing Labels data."""
    shape_a = (10, 15)
    shape_b = (20, 12)
    shape_c = (10, 10)
    np.random.seed(0)
    data_a = np.random.randint(20, size=shape_a)
    data_b = np.random.randint(20, size=shape_b)
    layer = Labels(data_a)
    layer.data = data_b
    np.testing.assert_array_equal(layer.data, data_b)
    assert layer.ndim == len(shape_b)
    np.testing.assert_array_equal(
        layer.extent.data[1], [s - 1 for s in shape_b]
    )
    assert layer._data_view.shape == shape_b[-2:]

    data_c = np.zeros(shape_c, dtype=bool)
    layer.data = data_c
    assert np.issubdtype(layer.data.dtype, np.integer)

    data_c = data_c.astype(np.float32)
    with pytest.raises(TypeError):
        layer.data = data_c


def test_changing_labels_dims():
    """Test changing Labels data including dimensionality."""
    shape_a = (10, 15)
    shape_b = (20, 12, 6)
    np.random.seed(0)
    data_a = np.random.randint(20, size=shape_a)
    data_b = np.random.randint(20, size=shape_b)
    layer = Labels(data_a)

    layer.data = data_b
    np.testing.assert_array_equal(layer.data, data_b)
    assert layer.ndim == len(shape_b)
    np.testing.assert_array_equal(
        layer.extent.data[1], [s - 1 for s in shape_b]
    )
    assert layer._data_view.shape == shape_b[-2:]


def test_changing_modes():
    """Test changing modes."""
    np.random.seed(0)
    data = np.random.randint(20, size=(10, 15))
    layer = Labels(data)
    assert layer.mode == 'pan_zoom'
    assert layer.mouse_pan is True

    layer.mode = 'fill'
    assert layer.mode == 'fill'
    assert layer.mouse_pan is False

    layer.mode = 'paint'
    assert layer.mode == 'paint'
    assert layer.mouse_pan is False

    layer.mode = 'pick'
    assert layer.mode == 'pick'
    assert layer.mouse_pan is False

    layer.mode = 'polygon'
    assert layer.mode == 'polygon'
    assert layer.mouse_pan is False

    layer.mode = 'pan_zoom'
    assert layer.mode == 'pan_zoom'
    assert layer.mouse_pan is True

    layer.mode = 'paint'
    assert layer.mode == 'paint'
    layer.editable = False
    assert layer.mode == 'pan_zoom'
    assert layer.editable is False


def test_name():
    """Test setting layer name."""
    np.random.seed(0)
    data = np.random.randint(20, size=(10, 15))
    layer = Labels(data)
    assert layer.name == 'Labels'

    layer = Labels(data, name='random')
    assert layer.name == 'random'

    layer.name = 'lbls'
    assert layer.name == 'lbls'


def test_visiblity():
    """Test setting layer visibility."""
    np.random.seed(0)
    data = np.random.randint(20, size=(10, 15))
    layer = Labels(data)
    assert layer.visible is True

    layer.visible = False
    assert layer.visible is False

    layer = Labels(data, visible=False)
    assert layer.visible is False

    layer.visible = True
    assert layer.visible is True


def test_opacity():
    """Test setting layer opacity."""
    np.random.seed(0)
    data = np.random.randint(20, size=(10, 15))
    layer = Labels(data)
    assert layer.opacity == 0.7

    layer.opacity = 0.5
    assert layer.opacity == 0.5

    layer = Labels(data, opacity=0.6)
    assert layer.opacity == 0.6

    layer.opacity = 0.3
    assert layer.opacity == 0.3


def test_blending():
    """Test setting layer blending."""
    np.random.seed(0)
    data = np.random.randint(20, size=(10, 15))
    layer = Labels(data)
    assert layer.blending == 'translucent'

    layer.blending = 'additive'
    assert layer.blending == 'additive'

    layer = Labels(data, blending='additive')
    assert layer.blending == 'additive'

    layer.blending = 'opaque'
    assert layer.blending == 'opaque'


@pytest.mark.filterwarnings('ignore:.*seed is deprecated.*')
def test_properties():
    """Test adding labels with properties."""
    np.random.seed(0)
    data = np.random.randint(20, size=(10, 15))

    layer = Labels(data)
    assert isinstance(layer.properties, dict)
    assert len(layer.properties) == 0

    properties = {
        'class': np.array(['Background'] + [f'Class {i}' for i in range(20)])
    }
    label_index = {i: i for i in range(len(properties['class']))}
    layer = Labels(data, properties=properties)
    assert isinstance(layer.properties, dict)
    np.testing.assert_equal(layer.properties, properties)
    assert layer._label_index == label_index
    layer = Labels(data)
    layer.properties = properties
    assert isinstance(layer.properties, dict)
    np.testing.assert_equal(layer.properties, properties)
    assert layer._label_index == label_index

    current_label = layer.get_value((0, 0))
    layer_message = layer.get_status((0, 0))
    assert layer_message['coordinates'].endswith(f'Class {current_label - 1}')

    properties = {'class': ['Background']}
    layer = Labels(data, properties=properties)
    layer_message = layer.get_status((0, 0))
    assert layer_message['coordinates'].endswith('[No Properties]')

    properties = {'class': ['Background', 'Class 12'], 'index': [0, 12]}
    label_index = {0: 0, 12: 1}
    layer = Labels(data, properties=properties)
    layer_message = layer.get_status((0, 0))
    assert layer._label_index == label_index
    assert layer_message['coordinates'].endswith('Class 12')

    layer = Labels(data)
    layer.properties = properties
    layer_message = layer.get_status((0, 0))
    assert layer._label_index == label_index
    assert layer_message['coordinates'].endswith('Class 12')

    layer = Labels(data)
    layer.properties = pd.DataFrame(properties)
    layer_message = layer.get_status((0, 0))
    assert layer._label_index == label_index
    assert layer_message['coordinates'].endswith('Class 12')


def test_default_properties_assignment():
    """Test that the default properties value can be assigned to properties
    see https://github.com/napari/napari/issues/2477
    """
    np.random.seed(0)
    data = np.random.randint(20, size=(10, 15))

    layer = Labels(data)
    layer.properties = {}
    assert layer.properties == {}


def test_multiscale_properties():
    """Test adding labels with multiscale properties."""
    np.random.seed(0)
    data0 = np.random.randint(20, size=(10, 15))
    data1 = data0[::2, ::2]
    data = [data0, data1]

    layer = Labels(data)
    assert isinstance(layer.properties, dict)
    assert len(layer.properties) == 0

    properties = {
        'class': np.array(['Background'] + [f'Class {i}' for i in range(20)])
    }
    label_index = {i: i for i in range(len(properties['class']))}
    layer = Labels(data, properties=properties)
    assert isinstance(layer.properties, dict)
    np.testing.assert_equal(layer.properties, properties)
    assert layer._label_index == label_index

    current_label = layer.get_value((0, 0))[1]
    layer_message = layer.get_status((0, 0))
    assert layer_message['coordinates'].endswith(f'Class {current_label - 1}')

    properties = {'class': ['Background']}
    layer = Labels(data, properties=properties)
    layer_message = layer.get_status((0, 0))
    assert layer_message['coordinates'].endswith('[No Properties]')

    properties = {'class': ['Background', 'Class 12'], 'index': [0, 12]}
    label_index = {0: 0, 12: 1}
    layer = Labels(data, properties=properties)
    layer_message = layer.get_status((0, 0))
    assert layer._label_index == label_index
    assert layer_message['coordinates'].endswith('Class 12')


def test_colormap():
    """Test colormap."""
    np.random.seed(0)
    data = np.random.randint(20, size=(10, 15))
    layer = Labels(data)
    assert isinstance(layer.colormap, Colormap)
    assert layer.colormap.name == 'label_colormap'

    layer.new_colormap()
    assert isinstance(layer.colormap, Colormap)
    assert layer.colormap.name == 'label_colormap'


def test_label_colormap():
    """Test a label colormap."""
    colormap = label_colormap(num_colors=4)

    # Make sure color 0 is transparent
    assert not np.any(colormap.map([0.0]))

    # test that all four colors are represented in a large set of random
    # labels.
    # we choose non-zero labels, and then there should not be any transparent
    # values.
    labels = np.random.randint(1, 2**23, size=(100, 100)).astype(np.float32)
    colormapped = colormap.map(labels)
    linear = np.reshape(colormapped, (-1, 4))
    unique = np.unique(linear, axis=0)
    assert len(unique) == 4


def test_custom_color_dict():
    """Test custom color dict."""
    np.random.seed(0)
    data = np.random.randint(20, size=(10, 15))
    cmap = DirectLabelColormap(
        color_dict=defaultdict(
            lambda: 'black',
            {2: 'white', 4: 'red', 8: 'blue', 16: 'red', 32: 'blue'},
        )
    )
    layer = Labels(data, colormap=cmap)

    # test with custom color dict
    assert isinstance(layer.get_color(2), np.ndarray)
    assert isinstance(layer.get_color(1), np.ndarray)
    assert (layer.get_color(2) == np.array([1.0, 1.0, 1.0, 1.0])).all()
    assert (layer.get_color(4) == layer.get_color(16)).all()
    assert (layer.get_color(8) == layer.get_color(32)).all()

    # test disable custom color dict
    # should not initialize as white since we are using random.seed
    assert not (layer.get_color(1) == np.array([1.0, 1.0, 1.0, 1.0])).all()


@pytest.mark.parametrize(
    'colormap_like',
    [
        ['red', 'blue'],
        [[1, 0, 0, 1], [0, 0, 1, 1]],
        {None: 'transparent', 1: 'red', 2: 'blue'},
        {None: [0, 0, 0, 0], 1: [1, 0, 0, 1], 2: [0, 0, 1, 1]},
        defaultdict(lambda: 'transparent', {1: 'red', 2: 'blue'}),
    ],
)
def test_colormap_simple_data_types(colormap_like):
    """Test that setting colormap with list or dict of colors works."""
    data = np.random.randint(20, size=(10, 15))
    # test in constructor
    _ = Labels(data, colormap=colormap_like)
    # test assignment
    layer = Labels(data)
    layer.colormap = colormap_like


def test_metadata():
    """Test setting labels metadata."""
    np.random.seed(0)
    data = np.random.randint(20, size=(10, 15))
    layer = Labels(data)
    assert layer.metadata == {}

    layer = Labels(data, metadata={'unit': 'cm'})
    assert layer.metadata == {'unit': 'cm'}


def test_brush_size():
    """Test changing brush size."""
    np.random.seed(0)
    data = np.random.randint(20, size=(10, 15))
    layer = Labels(data)
    assert layer.brush_size == 10

    layer.brush_size = 20
    assert layer.brush_size == 20


def test_contiguous():
    """Test changing contiguous."""
    np.random.seed(0)
    data = np.random.randint(20, size=(10, 15))
    layer = Labels(data)
    assert layer.contiguous is True

    layer.contiguous = False
    assert layer.contiguous is False


def test_n_edit_dimensions():
    """Test changing the number of editable dimensions."""
    np.random.seed(0)
    data = np.random.randint(20, size=(5, 10, 15))
    layer = Labels(data)
    layer.n_edit_dimensions = 2
    layer.n_edit_dimensions = 3


@pytest.mark.parametrize(
    ('input_data', 'expected_data_view'),
    [
        (
            np.array(
                [
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 5, 5, 5, 0, 0],
                    [0, 0, 1, 1, 1, 5, 5, 5, 0, 0],
                    [0, 0, 1, 1, 1, 5, 5, 5, 0, 0],
                    [0, 0, 1, 1, 1, 5, 5, 5, 0, 0],
                    [0, 0, 0, 0, 0, 5, 5, 5, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                ],
                dtype=np.int_,
            ),
            np.array(
                [
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 5, 5, 5, 0, 0],
                    [0, 0, 1, 1, 1, 5, 0, 5, 0, 0],
                    [0, 0, 1, 0, 1, 5, 0, 5, 0, 0],
                    [0, 0, 1, 1, 1, 5, 0, 5, 0, 0],
                    [0, 0, 0, 0, 0, 5, 5, 5, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                ],
                dtype=np.int_,
            ),
        ),
        (
            np.array(
                [
                    [1, 1, 0, 0, 0, 0, 0, 2, 2, 2],
                    [1, 1, 0, 0, 0, 0, 0, 2, 2, 2],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 4, 4, 4, 4],
                    [3, 3, 3, 0, 0, 0, 4, 4, 4, 4],
                    [3, 3, 3, 0, 0, 0, 4, 4, 4, 4],
                    [3, 3, 3, 0, 0, 0, 4, 4, 4, 4],
                ],
                dtype=np.int_,
            ),
            np.array(
                [
                    [0, 1, 0, 0, 0, 0, 0, 2, 0, 0],
                    [1, 1, 0, 0, 0, 0, 0, 2, 2, 2],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 4, 4, 4, 4],
                    [3, 3, 3, 0, 0, 0, 4, 0, 0, 0],
                    [0, 0, 3, 0, 0, 0, 4, 0, 0, 0],
                    [0, 0, 3, 0, 0, 0, 4, 0, 0, 0],
                ],
                dtype=np.int_,
            ),
        ),
        (
            5 * np.ones((9, 10), dtype=np.uint32),
            np.zeros((9, 10), dtype=np.uint32),
        ),
    ],
    ids=['touching objects', 'touching border', 'full array'],
)
def test_contour(input_data, expected_data_view):
    """Test changing contour."""
    layer = Labels(input_data)
    assert layer.contour == 0
    np.testing.assert_array_equal(layer.data, input_data)

    np.testing.assert_array_equal(
        layer._raw_to_displayed(input_data.astype(np.float32)),
        layer._data_view,
    )
    data_view_before_contour = layer._data_view.copy()

    layer.contour = 1
    assert layer.contour == 1

    # Check `layer.data` didn't change
    np.testing.assert_array_equal(layer.data, input_data)

    # Check what is returned in the view of the data
    np.testing.assert_array_equal(
        layer._data_view,
        np.where(
            expected_data_view > 0,
            expected_data_view,
            0,
        ),
    )

    # Check the view of the data changed after setting the contour
    with np.testing.assert_raises(AssertionError):
        np.testing.assert_array_equal(
            data_view_before_contour, layer._data_view
        )

    layer.contour = 0
    assert layer.contour == 0

    # Check it's in the same state as before setting the contour
    np.testing.assert_array_equal(
        layer._raw_to_displayed(input_data), layer._data_view
    )

    with pytest.raises(ValueError, match='contour value must be >= 0'):
        layer.contour = -1


@pytest.mark.parametrize('background_num', [0, 1, 2, -1])
def test_background_label(background_num):
    data = np.zeros((10, 10), dtype=np.int32)
    data[1:-1, 1:-1] = 1
    data[2:-2, 2:-2] = 2
    data[4:-4, 4:-4] = -1

    layer = Labels(data)
    layer.colormap = label_colormap(49, background_value=background_num)
    np.testing.assert_array_equal(
        layer._data_view == 0, data == background_num
    )
    np.testing.assert_array_equal(
        layer._data_view != 0, data != background_num
    )


def test_contour_large_new_labels():
    """Check that new labels larger than the lookup table work in contour mode.

    References
    ----------
    [1]: https://forum.image.sc/t/data-specific-reason-for-indexerror-in-raw-to-displayed/60808
    [2]: https://github.com/napari/napari/pull/3697
    """
    viewer = ViewerModel()

    labels = np.zeros((5, 10, 10), dtype=int)
    labels[0, 4:6, 4:6] = 1
    labels[4, 4:6, 4:6] = 1000
    labels_layer = viewer.add_labels(labels)
    labels_layer.contour = 1
    # This used to fail with IndexError
    viewer.dims.set_point(axis=0, value=4)


def test_contour_local_updates():
    """Checks if contours are rendered correctly with local updates"""
    data = np.zeros((7, 7), dtype=np.int32)

    layer = Labels(data)
    layer.contour = 1
    assert np.allclose(
        layer._raw_to_displayed(layer._slice.image.raw),
        np.zeros((7, 7), dtype=np.float32),
    )

    painting_mask = np.array(
        [
            [0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0],
            [0, 0, 1, 1, 1, 0, 0],
            [0, 0, 1, 1, 1, 0, 0],
            [0, 0, 1, 1, 1, 0, 0],
            [0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0],
        ],
        dtype=np.int32,
    )

    layer.data_setitem(np.nonzero(painting_mask), 1, refresh=True)

    assert np.array_equiv(
        (layer._slice.image.view > 0), get_contours(painting_mask, 1, 0)
    )


def test_data_setitem_multi_dim():
    """
    this test checks if data_setitem works when some of the indices are
    outside currently rendered slice
    """
    # create zarr zeros array in memory
    data = zarr.zeros((10, 10, 10), chunks=(5, 5, 5), dtype=np.uint32)
    labels = Labels(data)
    labels.data_setitem(
        (np.array([0, 1, 1]), np.array([1, 1, 2]), np.array([0, 0, 0])),
        [1, 2, 0],
    )


def test_data_setitiem_transposed_axes():
    data = np.zeros((10, 100), dtype=np.uint32)
    labels = Labels(data)
    dims = Dims(ndim=2, ndisplay=2, order=(1, 0))
    labels.data_setitem((np.array([9]), np.array([99])), 1)
    labels._slice_dims(dims)
    labels.data_setitem((np.array([9]), np.array([99])), 2)


def test_selecting_label():
    """Test selecting label."""
    np.random.seed(0)
    data = np.random.randint(20, size=(10, 15))
    layer = Labels(data)
    assert layer.selected_label == 1
    assert (layer._selected_color == layer.get_color(1)).all

    layer.selected_label = 1
    assert layer.selected_label == 1
    assert len(layer._selected_color) == 4


def test_label_color():
    """Test getting label color."""
    np.random.seed(0)
    data = np.random.randint(20, size=(10, 15))
    layer = Labels(data)
    col = layer.get_color(0)
    assert col is None

    col = layer.get_color(1)
    assert len(col) == 4


def test_show_selected_label():
    """Test color of labels when filtering to selected labels"""
    np.random.seed(0)
    data = np.random.randint(20, size=(10, 15))
    layer = Labels(data)
    original_color = layer.get_color(1)

    layer.show_selected_label = True
    original_background_color = layer.get_color(
        layer.colormap.background_value
    )
    none_color = layer.get_color(None)
    layer.selected_label = 1

    # color of selected label has not changed
    assert np.allclose(layer.get_color(layer.selected_label), original_color)

    current_background_color = layer.get_color(layer.colormap.background_value)
    # color of background is background color
    assert current_background_color == original_background_color

    # color of all others is none color
    other_labels = np.unique(layer.data)[2:]
    other_colors = np.array([layer.get_color(x) for x in other_labels])
    assert np.allclose(other_colors, none_color)


def test_paint():
    """Test painting labels with different circle brush sizes."""
    np.random.seed(0)
    data = np.random.randint(20, size=(10, 15))
    data[:10, :10] = 1
    layer = Labels(data)

    assert np.unique(layer.data[:5, :5]) == 1
    assert np.unique(layer.data[5:10, 5:10]) == 1

    layer.brush_size = 9
    layer.paint([0, 0], 2)
    assert np.unique(layer.data[:4, :4]) == 2
    assert np.unique(layer.data[5:10, 5:10]) == 1

    layer.brush_size = 10
    layer.paint([0, 0], 2)
    assert np.unique(layer.data[0:6, 0:3]) == 2
    assert np.unique(layer.data[0:3, 0:6]) == 2
    assert np.unique(layer.data[6:10, 6:10]) == 1

    layer.brush_size = 19
    layer.paint([0, 0], 2)
    assert np.unique(layer.data[0:4, 0:10]) == 2
    assert np.unique(layer.data[0:10, 0:4]) == 2
    assert np.unique(layer.data[3:7, 3:7]) == 2
    assert np.unique(layer.data[7:10, 7:10]) == 1


def test_paint_with_preserve_labels():
    """Test painting labels with square brush while preserving existing labels."""
    data = np.zeros((15, 10), dtype=np.uint32)
    data[:3, :3] = 1
    layer = Labels(data)

    layer.preserve_labels = True
    assert np.unique(layer.data[:3, :3]) == 1

    layer.brush_size = 9
    layer.paint([0, 0], 2)

    assert np.unique(layer.data[3:5, 0:3]) == 2
    assert np.unique(layer.data[0:3, 3:5]) == 2
    assert np.unique(layer.data[:3, :3]) == 1


def test_setting_prev_selected_label():
    """Test that _prev_selected_label is set when selected_label
    is set to the colormap background_value."""
    data = np.zeros((15, 10), dtype=np.uint32)
    layer = Labels(data)

    # set the selected label, _prev_selected_label should be None
    layer.selected_label = 1
    assert not layer._prev_selected_label

    layer.selected_label = layer.colormap.background_value
    assert layer._prev_selected_label == 1

    layer.selected_label = 2
    # swap the selected and background labels
    # _prev_selected_label should be set
    layer.swap_selected_and_background_labels()
    assert layer.selected_label == layer.colormap.background_value
    assert layer._prev_selected_label == 2


def test_paint_swap_with_preserve_labels():
    """Test painting and swapping labels & background while preserving labels."""
    data = np.zeros((15, 10), dtype=np.uint32)
    data[:3, :3] = 1
    layer = Labels(data)

    layer.preserve_labels = True
    assert np.unique(layer.data[:3, :3]) == 1

    layer.brush_size = 9
    layer.selected_label = 2
    layer.paint([0, 0], layer.selected_label)

    assert np.unique(layer.data[3:5, 0:3]) == 2
    assert np.unique(layer.data[0:3, 3:5]) == 2
    assert np.unique(layer.data[:3, :3]) == 1

    current_label = layer.selected_label
    layer.swap_selected_and_background_labels()
    assert layer.selected_label == 0
    assert layer._prev_selected_label == current_label
    layer.paint([0, 0], layer.selected_label)

    # only label 2 should be erased
    assert np.unique(layer.data[3:5, 0:3]) == 0
    assert np.unique(layer.data[0:3, 3:5]) == 0
    assert np.unique(layer.data[:3, :3]) == 1


def test_paint_2d():
    """Test painting labels with circle brush."""
    data = np.zeros((40, 40), dtype=np.uint32)
    layer = Labels(data)
    layer.brush_size = 12
    layer.mode = 'paint'
    layer.paint((0, 0), 3)

    layer.brush_size = 12
    layer.paint((15, 8), 4)

    layer.brush_size = 13
    layer.paint((30.2, 7.8), 5)

    layer.brush_size = 12
    layer.paint((39, 39), 6)

    layer.brush_size = 20
    layer.paint((15, 27), 7)

    assert np.sum(layer.data[:8, :8] == 3) == 41
    assert np.sum(layer.data[9:22, 2:15] == 4) == 137
    assert np.sum(layer.data[24:37, 2:15] == 5) == 137
    assert np.sum(layer.data[33:, 33:] == 6) == 41
    assert np.sum(layer.data[5:26, 17:38] == 7) == 349


def test_paint_2d_xarray():
    """Test the memory usage of painting a xarray indirectly via timeout."""
    now = time.monotonic()
    data = xr.DataArray(np.zeros((3, 3, 1024, 1024), dtype=np.uint32))

    layer = Labels(data)
    layer.brush_size = 12
    layer.mode = 'paint'
    layer.paint((1, 1, 512, 512), 3)
    assert isinstance(layer.data, xr.DataArray)
    assert layer.data.sum() == 411
    elapsed = time.monotonic() - now
    assert elapsed < 1, 'test was too slow, computation was likely not lazy'


def test_paint_3d():
    """Test painting labels with circle brush on 3D image."""
    data = np.zeros((30, 40, 40), dtype=np.uint32)
    layer = Labels(data)
    layer.brush_size = 12
    layer.mode = 'paint'

    # Paint in 2D
    layer.paint((10, 10, 10), 3)

    # Paint in 3D
    layer.n_edit_dimensions = 3
    layer.paint((10, 25, 10), 4)

    # Paint in 3D, preserve labels
    layer.n_edit_dimensions = 3
    layer.preserve_labels = True
    layer.paint((10, 15, 15), 5)

    assert np.sum(layer.data[4:17, 4:17, 4:17] == 3) == 137
    assert np.sum(layer.data[4:17, 19:32, 4:17] == 4) == 1189
    assert np.sum(layer.data[4:17, 9:32, 9:32] == 5) == 1103


def test_paint_polygon():
    """Test painting labels with polygons."""
    data = np.zeros((10, 15), dtype=int)
    data[:10, :10] = 1
    layer = Labels(data)

    layer.paint_polygon([[0, 0], [0, 5], [5, 5], [5, 0]], 2)
    assert np.array_equiv(layer.data[:5, :5], 2)
    assert np.array_equiv(layer.data[:10, 6:10], 1)
    assert np.array_equiv(layer.data[6:10, :10], 1)

    layer.paint_polygon([[7, 7], [7, 7], [7, 7]], 3)
    assert layer.data[7, 7] == 3
    assert np.array_equiv(
        layer.data[[6, 7, 8, 7, 8, 6], [7, 6, 7, 8, 8, 6]], 1
    )

    data[:10, :10] = 0
    gt_pattern = np.array(
        [
            [0, 0, 0, 0, 0, 0, 0, 0],
            [0, 1, 1, 1, 1, 1, 1, 0],
            [0, 1, 1, 1, 1, 1, 1, 0],
            [0, 1, 1, 0, 0, 1, 1, 0],
            [0, 1, 1, 0, 0, 1, 1, 0],
            [0, 1, 1, 0, 0, 1, 1, 0],
            [0, 0, 0, 0, 0, 0, 0, 0],
        ]
    )
    polygon_points = [
        [1, 1],
        [1, 6],
        [5, 6],
        [5, 5],
        [2, 5],
        [2, 2],
        [5, 2],
        [5, 1],
    ]
    layer.paint_polygon(polygon_points, 1)
    assert np.allclose(layer.data[:7, :8], gt_pattern)

    data[:10, :10] = 0
    layer.paint_polygon(polygon_points[::-1], 1)
    assert np.allclose(layer.data[:7, :8], gt_pattern)


def test_paint_polygon_2d_in_3d():
    """Test painting labels with polygons in a 3D array"""
    data = np.zeros((3, 10, 10), dtype=int)
    layer = Labels(data)

    assert layer.n_edit_dimensions == 2

    layer.paint_polygon([[1, 0, 0], [1, 0, 9], [1, 9, 9], [1, 9, 0]], 1)

    assert np.array_equiv(data[1, :], 1)
    assert np.array_equiv(data[[0, 2], :], 0)


def test_fill():
    """Test filling labels with different brush sizes."""
    np.random.seed(0)
    data = np.random.randint(20, size=(10, 15))
    data[:10, :10] = 2
    data[:5, :5] = 1
    layer = Labels(data)
    assert np.unique(layer.data[:5, :5]) == 1
    assert np.unique(layer.data[5:10, 5:10]) == 2

    layer.fill([0, 0], 3)
    assert np.unique(layer.data[:5, :5]) == 3
    assert np.unique(layer.data[5:10, 5:10]) == 2


def test_fill_swap_with_preserve_labels():
    """Test fill and swap background while preserving existing labels."""
    data = np.zeros((15, 10), dtype=np.uint32)
    data[:3, :3] = 1
    layer = Labels(data)

    layer.preserve_labels = True
    assert np.unique(layer.data[:3, :3]) == 1

    layer.selected_label = 2
    prev_layer_data = layer.data
    layer.fill([0, 0], layer.selected_label)
    # existing label should not be filled
    assert np.all(layer.data == prev_layer_data)

    layer.fill([3, 3], layer.selected_label)
    # background should be filled
    assert np.unique(layer.data[3:, 0:3]) == 2
    assert np.unique(layer.data[0:3, 3:]) == 2
    # existing label should not be filled
    assert np.unique(layer.data[:3, :3]) == 1

    current_label = layer.selected_label
    layer.swap_selected_and_background_labels()
    assert layer.selected_label == 0
    assert layer._prev_selected_label == current_label
    prev_layer_data = layer.data
    layer.fill([0, 0], layer.selected_label)
    # existing label should not be erased
    assert np.all(layer.data == prev_layer_data)

    layer.fill([3, 3], layer.selected_label)
    # the pre-swap label should be erassed
    assert np.unique(layer.data[3:, 0:3]) == 0
    assert np.unique(layer.data[0:3, 3:]) == 0
    # existing label should not be filled
    assert np.unique(layer.data[:3, :3]) == 1


def test_value():
    """Test getting the value of the data at the current coordinates."""
    np.random.seed(0)
    data = np.random.randint(20, size=(10, 15))
    layer = Labels(data)
    value = layer.get_value((0, 0))
    assert value == data[0, 0]


@pytest.mark.parametrize(
    ('position', 'view_direction', 'dims_displayed', 'world'),
    [
        ([10, 5, 5], [1, 0, 0], [0, 1, 2], False),
        ([10, 5, 5], [1, 0, 0], [0, 1, 2], True),
        ([0, 10, 5, 5], [0, 1, 0, 0], [1, 2, 3], True),
    ],
)
def test_value_3d(position, view_direction, dims_displayed, world):
    """get_value should return label value in 3D"""
    data = np.zeros((20, 20, 20), dtype=int)
    data[0:10, 0:10, 0:10] = 1
    layer = Labels(data)
    layer._slice_dims(Dims(ndim=3, ndisplay=3))
    value = layer.get_value(
        position,
        view_direction=view_direction,
        dims_displayed=dims_displayed,
        world=world,
    )
    assert value == 1


def test_message():
    """Test converting value and coords to message."""
    np.random.seed(0)
    data = np.random.randint(20, size=(10, 15))
    layer = Labels(data)
    msg = layer.get_status((0, 0))
    assert isinstance(msg, dict)


def test_thumbnail():
    """Test the image thumbnail for square data."""
    np.random.seed(0)
    data = np.random.randint(20, size=(30, 30))
    layer = Labels(data)
    layer._update_thumbnail()
    assert layer.thumbnail.shape == layer._thumbnail_shape


@pytest.mark.parametrize('value', [1, 10, 50, -2, -10])
@pytest.mark.parametrize('dtype', [np.int8, np.int32])
def test_thumbnail_single_color(value, dtype):
    labels = Labels(np.full((10, 10), value, dtype=dtype), opacity=1)
    labels._update_thumbnail()
    mid_point = tuple(np.array(labels.thumbnail.shape[:2]) // 2)
    npt.assert_array_equal(
        labels.thumbnail[mid_point], labels.get_color(value) * 255
    )


def test_world_data_extent():
    """Test extent after applying transforms."""
    np.random.seed(0)
    shape = (6, 10, 15)
    data = np.random.randint(20, size=shape)
    layer = Labels(data)
    extent = np.array(((0,) * 3, [s - 1 for s in shape]))
    check_layer_world_data_extent(layer, extent, (3, 1, 1), (10, 20, 5))


@pytest.mark.parametrize(
    (
        'brush_size',
        'mode',
        'selected_label',
        'preserve_labels',
        'n_edit_dimensions',
    ),
    list(
        itertools.product(
            list(range(1, 22, 5)),
            ['fill', 'erase', 'paint'],
            [1, 20, 100],
            [True, False],
            [3, 2],
        )
    ),
)
def test_undo_redo(
    brush_size,
    mode,
    selected_label,
    preserve_labels,
    n_edit_dimensions,
):
    blobs = sk_data.binary_blobs(length=64, volume_fraction=0.3, n_dim=3)
    layer = Labels(blobs)
    data_history = [blobs.copy()]
    layer.brush_size = brush_size
    layer.mode = mode
    layer.selected_label = selected_label
    layer.preserve_labels = preserve_labels
    layer.n_edit_dimensions = n_edit_dimensions
    coord = np.random.random((3,)) * (np.array(blobs.shape) - 1)
    while layer.data[tuple(coord.astype(int))] == 0 and np.any(layer.data):
        coord = np.random.random((3,)) * (np.array(blobs.shape) - 1)
    if layer.mode == 'fill':
        layer.fill(coord, layer.selected_label)
    if layer.mode == 'erase':
        layer.paint(coord, 0)
    if layer.mode == 'paint':
        layer.paint(coord, layer.selected_label)
    data_history.append(np.copy(layer.data))
    layer.undo()
    np.testing.assert_array_equal(layer.data, data_history[0])
    layer.redo()
    np.testing.assert_array_equal(layer.data, data_history[1])


def test_ndim_fill():
    test_array = np.zeros((5, 5, 5, 5), dtype=int)

    test_array[:, 1:3, 1:3, 1:3] = 1

    layer = Labels(test_array)
    layer.n_edit_dimensions = 3

    layer.fill((0, 1, 1, 1), 2)

    np.testing.assert_equal(layer.data[0, 1:3, 1:3, 1:3], 2)
    np.testing.assert_equal(layer.data[1, 1:3, 1:3, 1:3], 1)

    layer.n_edit_dimensions = 4

    layer.fill((1, 1, 1, 1), 3)

    np.testing.assert_equal(layer.data[0, 1:3, 1:3, 1:3], 2)
    np.testing.assert_equal(layer.data[1:, 1:3, 1:3, 1:3], 3)


def test_ndim_paint():
    test_array = np.zeros((5, 6, 7, 8), dtype=int)
    layer = Labels(test_array)
    layer.n_edit_dimensions = 3
    layer.brush_size = 2  # equivalent to 18-connected 3D neighborhood
    layer.paint((1, 1, 1, 1), 1)

    assert np.sum(layer.data) == 19  # 18 + center
    assert not np.any(layer.data[0])
    assert not np.any(layer.data[2:])

    layer.n_edit_dimensions = 2  # 3x3 square
    layer._slice_dims(Dims(ndim=4, order=(1, 2, 0, 3)))
    layer.paint((4, 5, 6, 7), 8)
    assert len(np.flatnonzero(layer.data == 8)) == 4  # 2D square is in corner
    np.testing.assert_array_equal(
        test_array[:, 5, 6, :],
        np.array(
            [
                [0, 0, 0, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 0, 8, 8],
                [0, 0, 0, 0, 0, 0, 8, 8],
            ]
        ),
    )


def test_cursor_size_with_negative_scale():
    layer = Labels(np.zeros((5, 5), dtype=int), scale=[-1, -1])
    layer.mode = 'paint'
    assert layer.cursor_size > 0


def test_large_label_values():
    label_array = 2**23 + np.arange(4, dtype=np.uint64).reshape((2, 2))
    layer = Labels(label_array)
    mapped = layer._random_colormap.map(layer.data)

    assert len(np.unique(mapped.reshape((-1, 4)), axis=0)) == 4


if parse_version(version('zarr')) > parse_version('3.0.0a0'):
    driver = [(2, 'zarr'), (3, 'zarr3')]
else:
    driver = [(2, 'zarr')]


@pytest.mark.parametrize(('zarr_version', 'zarr_driver'), driver)
def test_fill_tensorstore(tmp_path, zarr_version, zarr_driver):
    ts = pytest.importorskip('tensorstore')

    labels = np.zeros((5, 7, 8, 9), dtype=int)
    labels[1, 2:4, 4:6, 4:6] = 1
    labels[1, 3:5, 5:7, 6:8] = 2
    labels[2, 3:5, 5:7, 6:8] = 3

    file_path = str(tmp_path / 'labels.zarr')

    labels_temp = zarr.open(
        store=file_path,
        mode='w',
        shape=labels.shape,
        dtype=np.uint32,
        chunks=(1, 1, 8, 9),
        zarr_version=zarr_version,
    )
    labels_temp[:] = labels
    labels_ts_spec = {
        'driver': zarr_driver,
        'kvstore': {'driver': 'file', 'path': file_path},
        'path': '',
    }
    data = ts.open(labels_ts_spec, create=False, open=True).result()
    layer = Labels(data)
    layer.n_edit_dimensions = 3
    layer.fill((1, 4, 6, 7), 4)
    modified_labels = np.where(labels == 2, 4, labels)
    np.testing.assert_array_equal(modified_labels, np.asarray(data))


def test_fill_with_xarray():
    """See https://github.com/napari/napari/issues/2374"""
    data = xr.DataArray(np.zeros((5, 4, 4), dtype=int))
    layer = Labels(data)

    layer.fill((0, 2, 2), 1)

    np.testing.assert_array_equal(layer.data[0, :, :], np.ones((4, 4)))
    np.testing.assert_array_equal(layer.data[1:, :, :], np.zeros((4, 4, 4)))
    # In the associated issue, using xarray.DataArray caused memory allocation
    # problems due to different read indexing rules, so check that the data
    # saved for undo has the expected vectorized shape and values.
    undo_data = layer._undo_history[0][0][1]
    np.testing.assert_array_equal(undo_data, np.zeros((16,)))


@pytest.mark.parametrize(
    'scale', list(itertools.product([-2, 2], [-0.5, 0.5], [-0.5, 0.5]))
)
def test_paint_3d_negative_scale(scale):
    labels = np.zeros((3, 5, 11, 11), dtype=int)
    labels_layer = Labels(
        labels, scale=(1, *scale), translate=(-200, 100, 100)
    )
    labels_layer.n_edit_dimensions = 3
    labels_layer.brush_size = 8
    labels_layer.paint((1, 2, 5, 5), 1)
    np.testing.assert_array_equal(
        np.sum(labels_layer.data, axis=(1, 2, 3)), [0, 95, 0]
    )


def test_rendering_init():
    shape = (6, 10, 15)
    np.random.seed(0)
    data = np.random.randint(20, size=shape)
    layer = Labels(data, rendering='iso_categorical')

    assert layer.rendering == LabelsRendering.ISO_CATEGORICAL.value


def test_3d_video_and_3d_scale_translate_then_scale_translate_padded():
    # See the GitHub issue for more details:
    # https://github.com/napari/napari/issues/2967
    data = np.zeros((3, 5, 11, 11), dtype=int)
    labels = Labels(data, scale=(2, 1, 1), translate=(5, 5, 5))

    np.testing.assert_array_equal(labels.scale, (1, 2, 1, 1))
    np.testing.assert_array_equal(labels.translate, (0, 5, 5, 5))


@dataclass
class MouseEvent:
    # mock mouse event class
    pos: list[int]
    position: list[int]
    dims_point: list[int]
    dims_displayed: list[int]
    view_direction: list[int]


def test_get_value_ray_3d():
    """Test using _get_value_ray to interrogate labels in 3D"""
    # make a mock mouse event
    mouse_event = MouseEvent(
        pos=[25, 25],
        position=[10, 5, 5],
        dims_point=[1, 0, 0, 0],
        dims_displayed=[1, 2, 3],
        view_direction=[1, 0, 0],
    )
    data = np.zeros((5, 20, 20, 20), dtype=int)
    data[1, 0:10, 0:10, 0:10] = 1
    labels = Labels(data, scale=(1, 2, 1, 1), translate=(5, 5, 5))

    # set the dims to the slice with labels
    labels._slice_dims(Dims(ndim=4, ndisplay=3, point=(1, 0, 0, 0)))

    value = labels._get_value_ray(
        start_point=np.array([1, 0, 5, 5]),
        end_point=np.array([1, 20, 5, 5]),
        dims_displayed=mouse_event.dims_displayed,
    )
    assert value == 1

    # check with a ray that only goes through background
    value = labels._get_value_ray(
        start_point=np.array([1, 0, 15, 15]),
        end_point=np.array([1, 20, 15, 15]),
        dims_displayed=mouse_event.dims_displayed,
    )
    assert value is None

    # set the dims to a slice without labels
    labels._slice_dims(Dims(ndim=4, ndisplay=3, point=(0, 0, 0, 0)))

    value = labels._get_value_ray(
        start_point=np.array([0, 0, 5, 5]),
        end_point=np.array([0, 20, 5, 5]),
        dims_displayed=mouse_event.dims_displayed,
    )
    assert value is None


def test_get_value_ray_3d_rolled():
    """Test using _get_value_ray to interrogate labels in 3D
    with the dimensions rolled.
    """
    # make a mock mouse event
    mouse_event = MouseEvent(
        pos=[25, 25],
        position=[10, 5, 5, 1],
        dims_point=[0, 0, 0, 1],
        dims_displayed=[0, 1, 2],
        view_direction=[1, 0, 0, 0],
    )
    data = np.zeros((20, 20, 20, 5), dtype=int)
    data[0:10, 0:10, 0:10, 1] = 1
    labels = Labels(data, scale=(1, 2, 1, 1), translate=(5, 5, 5, 0))

    # set the dims to the slice with labels
    labels._slice_dims(
        Dims(ndim=4, ndisplay=3, order=(3, 0, 1, 2), point=(0, 0, 0, 1))
    )
    labels.set_view_slice()

    value = labels._get_value_ray(
        start_point=np.array([0, 5, 5, 1]),
        end_point=np.array([20, 5, 5, 1]),
        dims_displayed=mouse_event.dims_displayed,
    )
    assert value == 1


def test_get_value_ray_3d_transposed():
    """Test using _get_value_ray to interrogate labels in 3D
    with the dimensions transposed.
    """
    # make a mock mouse event
    mouse_event = MouseEvent(
        pos=[25, 25],
        position=[10, 5, 5, 1],
        dims_point=[0, 0, 0, 1],
        dims_displayed=[1, 3, 2],
        view_direction=[1, 0, 0, 0],
    )
    data = np.zeros((5, 20, 20, 20), dtype=int)
    data[1, 0:10, 0:10, 0:10] = 1
    labels = Labels(data, scale=(1, 2, 1, 1), translate=(0, 5, 5, 5))

    # set the dims to the slice with labels
    labels._slice_dims(
        Dims(ndim=4, ndisplay=3, order=(0, 1, 3, 2), point=(1, 0, 0, 0))
    )
    labels.set_view_slice()

    value = labels._get_value_ray(
        start_point=np.array([1, 0, 5, 5]),
        end_point=np.array([1, 20, 5, 5]),
        dims_displayed=mouse_event.dims_displayed,
    )
    assert value == 1


def test_get_value_ray_2d():
    """_get_value_ray currently only returns None in 2D
    (i.e., it shouldn't be used for 2D).
    """
    # make a mock mouse event
    mouse_event = MouseEvent(
        pos=[25, 25],
        position=[5, 5],
        dims_point=[1, 10, 0, 0],
        dims_displayed=[2, 3],
        view_direction=[1, 0, 0],
    )
    data = np.zeros((5, 20, 20, 20), dtype=int)
    data[1, 0:10, 0:10, 0:10] = 1
    labels = Labels(data, scale=(1, 2, 1, 1), translate=(5, 5, 5))

    # set the dims to the slice with labels, but 2D
    labels._slice_dims(Dims(ndim=4, ndisplay=2, point=(1, 10, 0, 0)))

    value = labels._get_value_ray(
        start_point=np.empty([]),
        end_point=np.empty([]),
        dims_displayed=mouse_event.dims_displayed,
    )
    assert value is None


def test_cursor_ray_3d():
    # make a mock mouse event
    mouse_event_1 = MouseEvent(
        pos=[25, 25],
        position=[1, 10, 27, 10],
        dims_point=[1, 0, 0, 0],
        dims_displayed=[1, 2, 3],
        view_direction=[0, 1, 0, 0],
    )
    data = np.zeros((5, 20, 20, 20), dtype=int)
    data[1, 0:10, 0:10, 0:10] = 1
    labels = Labels(data, scale=(1, 1, 2, 1), translate=(5, 5, 5))

    # set the slice to one with data and the view to 3D
    labels._slice_dims(Dims(ndim=4, ndisplay=3, point=(1, 0, 0, 0)))

    # axis 0 : [0, 20], bounding box extents along view axis, [1, 0, 0]
    # click is transformed: (value - translation) / scale
    # axis 1: click at 27 in world coords -> (27 - 5) / 2 = 11
    # axis 2: click at 10 in world coords -> (10 - 5) / 1 = 5
    start_point, end_point = labels.get_ray_intersections(
        mouse_event_1.position,
        mouse_event_1.view_direction,
        mouse_event_1.dims_displayed,
    )
    np.testing.assert_allclose(start_point, [1, 0, 11, 5])
    np.testing.assert_allclose(end_point, [1, 20, 11, 5])

    # click in the background
    mouse_event_2 = MouseEvent(
        pos=[25, 25],
        position=[1, 10, 65, 10],
        dims_point=[1, 0, 0, 0],
        dims_displayed=[1, 2, 3],
        view_direction=[0, 1, 0, 0],
    )
    start_point, end_point = labels.get_ray_intersections(
        mouse_event_2.position,
        mouse_event_2.view_direction,
        mouse_event_2.dims_displayed,
    )
    assert start_point is None
    assert end_point is None

    # click in a slice with no labels
    mouse_event_3 = MouseEvent(
        pos=[25, 25],
        position=[0, 10, 27, 10],
        dims_point=[0, 0, 0, 0],
        dims_displayed=[1, 2, 3],
        view_direction=[0, 1, 0, 0],
    )
    labels._slice_dims(Dims(ndim=4, ndisplay=3))
    start_point, end_point = labels.get_ray_intersections(
        mouse_event_3.position,
        mouse_event_3.view_direction,
        mouse_event_3.dims_displayed,
    )
    np.testing.assert_allclose(start_point, [0, 0, 11, 5])
    np.testing.assert_allclose(end_point, [0, 20, 11, 5])


def test_cursor_ray_3d_rolled():
    """Test that the cursor works when the displayed
    viewer axes have been rolled
    """
    # make a mock mouse event
    mouse_event_1 = MouseEvent(
        pos=[25, 25],
        position=[10, 27, 10, 1],
        dims_point=[0, 0, 0, 1],
        dims_displayed=[0, 1, 2],
        view_direction=[1, 0, 0, 0],
    )
    data = np.zeros((20, 20, 20, 5), dtype=int)
    data[0:10, 0:10, 0:10, 1] = 1
    labels = Labels(data, scale=(1, 2, 1, 1), translate=(5, 5, 5, 0))

    # set the slice to one with data and the view to 3D
    labels._slice_dims(Dims(ndim=4, ndisplay=3, point=(0, 0, 0, 1)))

    start_point, end_point = labels.get_ray_intersections(
        mouse_event_1.position,
        mouse_event_1.view_direction,
        mouse_event_1.dims_displayed,
    )
    np.testing.assert_allclose(start_point, [0, 11, 5, 1])
    np.testing.assert_allclose(end_point, [20, 11, 5, 1])


def test_cursor_ray_3d_transposed():
    """Test that the cursor works when the displayed
    viewer axes have been transposed
    """
    # make a mock mouse event
    mouse_event_1 = MouseEvent(
        pos=[25, 25],
        position=[10, 27, 10, 1],
        dims_point=[0, 0, 0, 1],
        dims_displayed=[0, 2, 1],
        view_direction=[1, 0, 0, 0],
    )
    data = np.zeros((20, 20, 20, 5), dtype=int)
    data[0:10, 0:10, 0:10, 1] = 1
    labels = Labels(data, scale=(1, 2, 1, 1), translate=(5, 5, 5, 0))

    # set the slice to one with data and the view to 3D
    labels._slice_dims(Dims(ndim=4, ndisplay=3, point=(0, 0, 0, 1)))

    start_point, end_point = labels.get_ray_intersections(
        mouse_event_1.position,
        mouse_event_1.view_direction,
        mouse_event_1.dims_displayed,
    )
    np.testing.assert_allclose(start_point, [0, 11, 5, 1])
    np.testing.assert_allclose(end_point, [20, 11, 5, 1])


def test_labels_state_update():
    """Test that a labels layer can be updated from the output of its
    _get_state() method
    """
    data = np.random.randint(20, size=(10, 15))
    layer = Labels(data)
    state = layer._get_state()
    for k, v in state.items():
        setattr(layer, k, v)


def test_is_default_color():
    """Test labels layer default color for None and background

    Previously, setting color to just default values would
    change color mode to DIRECT and display a black layer.
    This test ensures `is_default_color` is
    correctly checking against layer defaults, and `color_mode`
    is only changed when appropriate.

    See
        - https://github.com/napari/napari/issues/2479
        - https://github.com/napari/napari/issues/2953
    """
    data = np.random.randint(20, size=(10, 15))
    layer = Labels(data)

    # layer gets instantiated with defaults
    current_color = layer._direct_colormap.color_dict
    assert layer._is_default_colors(current_color)

    # setting color to default colors doesn't update color mode
    layer.colormap = DirectLabelColormap(color_dict=current_color)
    assert isinstance(layer.colormap, CyclicLabelColormap)

    # new colors are not default
    new_color = {0: 'white', 1: 'red', 3: 'green', None: 'blue'}
    assert not layer._is_default_colors(new_color)
    # setting the color with non-default colors updates color mode
    layer.colormap = DirectLabelColormap(color_dict=new_color)
    assert isinstance(layer.colormap, DirectLabelColormap)


def test_large_labels_direct_color():
    """Make sure direct color works with large label ranges"""
    pytest.importorskip('numba')
    data = np.array([[0, 1], [2**16, 2**20]], dtype=np.uint32)
    colors = {1: 'white', 2**16: 'green', 2**20: 'magenta'}
    layer = Labels(
        data,
        colormap=DirectLabelColormap(
            color_dict=defaultdict(lambda: 'black', colors)
        ),
    )
    np.testing.assert_allclose(layer.get_color(2**20), [1.0, 0.0, 1.0, 1.0])


def test_invalidate_cache_when_change_color_mode(
    direct_colormap, random_colormap
):
    """Checks if the cache is invalidated when color mode is changed."""
    data = np.zeros((4, 10), dtype=np.int32)
    data[1, :] = np.arange(0, 10)

    layer = Labels(data)
    layer.selected_label = 0
    gt_auto = layer._raw_to_displayed(layer._slice.image.raw)
    assert gt_auto.dtype == np.uint8

    layer.colormap = direct_colormap
    layer._cached_labels = None
    assert layer._raw_to_displayed(layer._slice.image.raw).dtype == np.uint8

    layer.colormap = random_colormap
    # If the cache is not invalidated, it returns colors for
    # the direct color mode instead of the color for the auto mode
    assert np.allclose(
        layer._raw_to_displayed(layer._slice.image.raw), gt_auto
    )


def test_color_mapping_when_color_is_changed():
    """Checks if the color mapping is computed correctly when the color palette is changed."""

    data = np.zeros((4, 5), dtype=np.int32)
    data[1, :] = np.arange(0, 5)
    layer = Labels(
        data,
        colormap=DirectLabelColormap(
            color_dict={1: 'green', 2: 'red', 3: 'white', None: 'black'}
        ),
    )
    gt_direct_3colors = layer._raw_to_displayed(layer._slice.image.raw)

    layer = Labels(
        data,
        colormap=DirectLabelColormap(
            color_dict={1: 'green', 2: 'red', None: 'black'}
        ),
    )
    assert layer._raw_to_displayed(layer._slice.image.raw).dtype == np.uint8
    layer.colormap = DirectLabelColormap(
        color_dict={1: 'green', 2: 'red', 3: 'white', None: 'black'}
    )

    assert np.allclose(
        layer._raw_to_displayed(layer._slice.image.raw), gt_direct_3colors
    )


def test_color_mapping_with_show_selected_label():
    """Checks if the color mapping is computed correctly when show_selected_label is activated."""

    data = np.arange(5, dtype=np.int32)[:, np.newaxis].repeat(5, axis=1)
    layer = Labels(data)
    mapped_colors_all = layer.colormap.map(data)

    layer.show_selected_label = True

    for selected_label in range(5):
        layer.selected_label = selected_label
        label_mask = data == selected_label
        mapped_colors = layer.colormap.map(data)

        npt.assert_allclose(
            mapped_colors[label_mask], mapped_colors_all[label_mask]
        )
        npt.assert_allclose(mapped_colors[np.logical_not(label_mask)], 0)

    layer.show_selected_label = False
    assert np.allclose(layer.colormap.map(data), mapped_colors_all)


def test_color_mapping_when_seed_is_changed():
    """Checks if the color mapping is updated when the color palette seed is changed."""
    np.random.seed(0)
    layer = Labels(np.random.randint(50, size=(10, 10)))
    mapped_colors1 = layer.colormap.map(
        layer._to_vispy_texture_dtype(layer._slice.image.raw)
    )

    layer.new_colormap()
    mapped_colors2 = layer.colormap.map(
        layer._to_vispy_texture_dtype(layer._slice.image.raw)
    )

    assert not np.allclose(mapped_colors1, mapped_colors2)


@pytest.mark.parametrize('num_colors', [49, 50, 254, 255, 60000, 65534])
def test_color_shuffling_above_num_colors(num_colors):
    r"""Check that the color shuffle does not result in the same collisions.

    See https://github.com/napari/napari/issues/6448.

    Note that we don't support more than 2\ :sup:`16` colors, and behavior
    with more colors is undefined, so we don't test it here.
    """
    labels = np.arange(1, 1 + 2 * (num_colors - 1)).reshape((2, -1))
    layer = Labels(labels, colormap=label_colormap(num_colors - 1))
    colors0 = layer.colormap.map(labels)
    assert np.all(colors0[0] == colors0[1])
    layer.new_colormap()
    colors1 = layer.colormap.map(labels)
    assert not np.all(colors1[0] == colors1[1])


def test_negative_label():
    """Test negative label values are supported."""
    data = np.random.randint(low=-1, high=20, size=(10, 10))
    original_data = np.copy(data)
    layer = Labels(data)
    layer.selected_label = -1
    layer.brush_size = 3
    layer.paint((5, 5), -1)
    assert np.count_nonzero(layer.data == -1) > np.count_nonzero(
        original_data == -1
    )


def test_negative_label_slicing():
    """Test negative label color doesn't change during slicing."""
    data = np.array([[[0, 1], [-1, -1]], [[100, 100], [-1, -2]]])
    layer = Labels(data)
    assert tuple(layer.get_color(1)) != tuple(layer.get_color(-1))
    layer._slice_dims(Dims(ndim=3, point=(1, 0, 0)))
    assert tuple(layer.get_color(-1)) != tuple(layer.get_color(100))
    assert tuple(layer.get_color(-2)) != tuple(layer.get_color(100))


def test_negative_label_doesnt_flicker():
    data = np.array(
        [
            [[0, 5], [0, 5]],
            [[-1, 5], [-1, 5]],
            [[-1, 6], [-1, 6]],
        ]
    )
    layer = Labels(data)
    layer._slice_dims(Dims(ndim=3, point=(1, 0, 0)))
    # This used to fail when negative values were used to index into _all_vals.
    assert tuple(layer.get_color(-1)) != tuple(layer.get_color(5))
    minus_one_color_original = tuple(layer.get_color(-1))
    layer.dims_point = (2, 0, 0)
    layer._set_view_slice()

    assert tuple(layer.get_color(-1)) == minus_one_color_original


def test_get_status_with_custom_index():
    """See https://github.com/napari/napari/issues/3811"""
    data = np.zeros((10, 10), dtype=np.uint8)
    data[2:5, 2:-2] = 1
    data[5:-2, 2:-2] = 2
    layer = Labels(data)
    df = pd.DataFrame(
        {'text1': [1, 3], 'text2': [7, -2], 'index': [1, 2]}, index=[1, 2]
    )
    layer.properties = df
    assert (
        layer.get_status((0, 0))['coordinates'] == ' [0 0]: 0; [No Properties]'
    )
    assert (
        layer.get_status((3, 3))['coordinates']
        == ' [3 3]: 1; text1: 1, text2: 7'
    )
    assert (
        layer.get_status((6, 6))['coordinates']
        == ' [6 6]: 2; text1: 3, text2: -2'
    )


def test_labels_features_event():
    event_emitted = False

    def on_event():
        nonlocal event_emitted
        event_emitted = True

    layer = Labels(np.zeros((4, 5), dtype=np.uint8))
    layer.events.features.connect(on_event)

    layer.features = {'some_feature': []}

    assert event_emitted


def test_copy():
    l1 = Labels(np.zeros((2, 4, 5), dtype=np.uint8))
    l2 = copy.copy(l1)
    l3 = Labels.create(*l1.as_layer_data_tuple())
    assert l1.data is not l2.data
    assert l1.data is l3.data


@pytest.mark.parametrize(
    ('colormap', 'expected'),
    [
        (label_colormap(49, 0.5), [0, 1]),
        (
            DirectLabelColormap(
                color_dict={
                    0: np.array([0, 0, 0, 0]),
                    1: np.array([1, 0, 0, 1]),
                    None: np.array([1, 1, 0, 1]),
                }
            ),
            [1, 2],
        ),
    ],
    ids=['auto', 'direct'],
)
def test_draw(colormap, expected):
    labels = Labels(np.zeros((30, 30), dtype=np.uint32))
    labels.mode = 'paint'
    labels.colormap = colormap
    labels.selected_label = 1
    npt.assert_array_equal(np.unique(labels._slice.image.raw), [0])
    npt.assert_array_equal(np.unique(labels._slice.image.view), expected[:1])
    labels._draw(1, (15, 15), (15, 15))
    npt.assert_array_equal(np.unique(labels._slice.image.raw), [0, 1])
    npt.assert_array_equal(np.unique(labels._slice.image.view), expected)


class TestLabels:
    @staticmethod
    def get_objects():
        return [(Labels(np.zeros((10, 10), dtype=np.uint8)))]

    def test_events_defined(self, event_define_check, obj):
        event_define_check(
            obj,
            {'seed', 'num_colors', 'color', 'seed_rng'},
        )


def test_docstring():
    validate_all_params_in_docstring(Labels)
    validate_kwargs_sorted(Labels)


def test_new_colormap_int8():
    """Check that int8 labels colors can be shuffled without overflow.

    See https://github.com/napari/napari/issues/7277.
    """
    data = np.arange(-128, 128, dtype=np.int8).reshape((16, 16))
    layer = Labels(data)
    layer.new_colormap(seed=0)


@pytest.mark.parametrize('visible', [True, False])
@pytest.mark.parametrize('dtype', [np.uint8, np.int8, np.uint32, np.int64])
def test_view_dtype(visible, dtype):
    layer = Labels(np.arange(25, dtype=dtype).reshape(5, 5), visible=visible)
    assert layer._slice.image.view.dtype == np.uint8


@pytest.mark.parametrize('visible', [True, False])
@pytest.mark.parametrize('dtype', [np.uint16, np.int16])
def test_view_dtype_int16(visible, dtype):
    layer = Labels(np.arange(25, dtype=dtype).reshape(5, 5), visible=visible)
    assert layer._slice.image.view.dtype == np.uint16



from napari.resources._icons import (
    ICON_PATH,
    ICONS,
    LOADING_GIF_PATH,
    get_colorized_svg,
    get_icon_path,
)

__all__ = [
    'ICONS',
    'ICON_PATH',
    'LOADING_GIF_PATH',
    'get_colorized_svg',
    'get_icon_path',
]



"""OpenGL Utilities."""

from collections.abc import Generator
from contextlib import contextmanager
from functools import lru_cache
from typing import Any, cast

import numpy as np
import numpy.typing as npt
from vispy.app import Canvas
from vispy.gloo import gl
from vispy.gloo.context import get_current_canvas

from napari.utils.translations import trans

texture_dtypes = [
    np.dtype(np.uint8),
    np.dtype(np.uint16),
    np.dtype(np.float32),
]


@contextmanager
def _opengl_context() -> Generator[None, None, None]:
    """Assure we are running with a valid OpenGL context.

    Only create a Canvas is one doesn't exist. Creating and closing a
    Canvas causes vispy to process Qt events which can cause problems.
    Ideally call opengl_context() on start after creating your first
    Canvas. However it will work either way.
    """
    canvas = Canvas(show=False) if get_current_canvas() is None else None
    try:
        yield
    finally:
        if canvas is not None:
            canvas.close()


@lru_cache(maxsize=1)
def get_gl_extensions() -> str:
    """Get basic info about the Gl capabilities of this machine"""
    with _opengl_context():
        return gl.glGetParameter(gl.GL_EXTENSIONS)


@lru_cache
def get_max_texture_sizes() -> tuple[int, int]:
    """Return the maximum texture sizes for 2D and 3D rendering.

    If this function is called without an OpenGL context it will create a
    temporary non-visible Canvas. Either way the lru_cache means subsequent
    calls to thing function will return the original values without
    actually running again.

    Returns
    -------
    Tuple[int, int]
        The max textures sizes for (2d, 3d) rendering.
    """
    with _opengl_context():
        max_size_2d = gl.glGetParameter(gl.GL_MAX_TEXTURE_SIZE)

    if max_size_2d == ():
        max_size_2d = None

    # vispy/gloo doesn't provide the GL_MAX_3D_TEXTURE_SIZE location,
    # but it can be found in this list of constants
    # http://pyopengl.sourceforge.net/documentation/pydoc/OpenGL.GL.html
    with _opengl_context():
        GL_MAX_3D_TEXTURE_SIZE = 32883
        max_size_3d = gl.glGetParameter(GL_MAX_3D_TEXTURE_SIZE)

    if max_size_3d == ():
        max_size_3d = None

    return max_size_2d, max_size_3d


def fix_data_dtype(data: npt.NDArray) -> npt.NDArray:
    """Makes sure the dtype of the data is accetpable to vispy.

    Acceptable types are int8, uint8, int16, uint16, float32.

    Parameters
    ----------
    data : np.ndarray
        Data that will need to be of right type.

    Returns
    -------
    np.ndarray
        Data that is of right type and will be passed to vispy.
    """

    dtype = np.dtype(data.dtype)
    if dtype in texture_dtypes:
        return data

    try:
        dtype_ = cast(
            'type[np.unsignedinteger[Any] | np.floating[Any]]',
            {
                'i': np.float32,
                'f': np.float32,
                'u': np.uint16,
                'b': np.uint8,
            }[dtype.kind],
        )
        if dtype_ == np.uint16 and dtype.itemsize > 2:
            dtype_ = np.float32
    except KeyError as e:  # not an int or float
        raise TypeError(
            trans._(
                'type {dtype} not allowed for texture; must be one of {textures}',
                deferred=True,
                dtype=dtype,
                textures=set(texture_dtypes),
            )
        ) from e
    return data.astype(dtype_)


# blend_func parameters are multiplying:
# - source color
# - destination color
# - source alpha
# - destination alpha
# they do not apply to min/max blending equation

BLENDING_MODES = {
    'opaque': {
        'depth_test': True,
        'cull_face': False,
        'blend': False,
    },
    'translucent': {
        'depth_test': True,
        'cull_face': False,
        'blend': True,
        'blend_func': ('src_alpha', 'one_minus_src_alpha', 'one', 'one'),
        'blend_equation': 'func_add',
    },
    'translucent_no_depth': {
        'depth_test': False,
        'cull_face': False,
        'blend': True,
        'blend_func': ('src_alpha', 'one_minus_src_alpha', 'one', 'one'),
        'blend_equation': 'func_add',  # see vispy/vispy#2324
    },
    'additive': {
        'depth_test': False,
        'cull_face': False,
        'blend': True,
        'blend_func': ('src_alpha', 'dst_alpha', 'one', 'one'),
        'blend_equation': 'func_add',
    },
    'minimum': {
        'depth_test': False,
        'cull_face': False,
        'blend': True,
        'blend_equation': 'min',
    },
    'multiplicative': {
        'depth_test': False,
        'cull_face': False,
        'blend': True,
        'blend_func': ('dst_color', 'zero', 'one', 'one'),
        'blend_equation': 'func_add',
    },
}



"""
This module is used to prevent a known issue with numpy<2 on macOS arm64
architecture installed from pypi wheels
(https://github.com/numpy/numpy/issues/21799).

We use a method to set thread limits based on the threadpoolctl package, but
reimplemented locally to prevent adding the dependency to napari.

Note: if any issues surface with the method below, we could fall back on
depending on threadpoolctl directly.

TODO: This module can be removed once the minimum numpy version is 2+.
"""

import ctypes
import logging
import os
import platform
from pathlib import Path

import numpy as np
from packaging.version import parse as parse_version

# if run with numpy<2 on macOS arm64 architecture compiled from pypi wheels,
# then it will crash with bus error if numpy is used in different thread
# Issue reported https://github.com/numpy/numpy/issues/21799
if (
    parse_version(np.__version__) < parse_version('2')
    and platform.system() == 'Darwin'
    and platform.machine() == 'arm64'
):  # pragma: no cover
    try:
        NUMPY_VERSION_IS_THREADSAFE = (
            'cibw-run'
            not in np.show_config('dicts')['Python Information']['path']
        )
    except (KeyError, TypeError):
        NUMPY_VERSION_IS_THREADSAFE = True
else:
    NUMPY_VERSION_IS_THREADSAFE = True


def limit_numpy1x_threads_on_macos_arm() -> (
    None
):  # pragma: no cover (macos only code)
    """Set openblas to use single thread on macOS arm64 to prevent numpy crash.

    On NumPy version<2 wheels on macOS ARM64 architectures, a BusError is
    raised, crashing Python, if NumPy is accessed from multiple threads.
    (See https://github.com/numpy/numpy/issues/21799.) This function uses the
    global check above (NUMPY_VERSION_IS_THREADSAFE), and, if False, it loads
    the linked OpenBLAS library and sets the number of threads to 1. This has
    performance implications but prevents nasty crashes, and anyway can be
    avoided by using more recent versions of NumPy.

    This function is loading openblas library from numpy and set number of threads to 1.
    See also:
        https://github.com/OpenMathLib/OpenBLAS/wiki/faq#how-can-i-use-openblas-in-multi-threaded-applications

    These changes seem to be sufficient to prevent the crashes.
    """
    if NUMPY_VERSION_IS_THREADSAFE:
        return
    # find openblas library
    numpy_dir = Path(np.__file__).parent
    if not (numpy_dir / '.dylibs').exists():
        logging.getLogger('napari').warning(
            'numpy .dylibs directory not found during try to prevent numpy crash'
        )
    # Recent numpy versions are built with cibuildwheel.
    # Internally, it uses delocate, which stores the openblas
    # library in the .dylibs directory.
    # Since we only patch numpy<2, we can just search for the libopenblas
    # dynamic library at this location.
    blas_lib = list((numpy_dir / '.dylibs').glob('libopenblas*.dylib'))
    if not blas_lib:
        logging.getLogger('napari').warning(
            'libopenblas not found during try to prevent numpy crash'
        )
        return

    blas = ctypes.CDLL(str(blas_lib[0]), mode=os.RTLD_NOLOAD)
    for suffix in ('', '64_', '_64'):
        openblas_set_num_threads = getattr(
            blas, f'openblas_set_num_threads{suffix}', None
        )
        if openblas_set_num_threads is not None:
            openblas_set_num_threads(1)
            break
    else:
        logging.getLogger('napari').warning(
            'openblas_set_num_threads not found'
        )
